/*
 * (c) Copyright 2003 Christian Lorenz - ALL RIGHTS RESERVED
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is part of the JavaBluetooth Stack.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * Created on Jul 6, 2003
 * by Christian Lorenz
 */

package javax.microedition.io;

import java.io.*;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import org.javabluetooth.stack.BluetoothStack;
import org.javabluetooth.stack.hci.HCIException;
import org.javabluetooth.stack.l2cap.JSR82Channel;
import org.javabluetooth.stack.l2cap.JSR82ConnectionNotifier;
import org.javabluetooth.stack.sco.SCOConnection;
import org.javabluetooth.stack.sco.SCOConnectionNotifier;
import org.javabluetooth.stack.sco.SCOException;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.BluetoothAddressException;
import org.javabluetooth.util.Debug;

/** 
 * This Class is a very basic implementation of the 
 * <code>javax.microedition.io.Connector</code> class. It should be used for 
 * Java Platforms that do not support the Java CDC, and only provides 
 * minimalistic features, required by <code>javax.bluetooth</code>.
 * If possible the system native of <code>javax.microedition.io.Connector</code> 
 * should be used. This class provides the static methods used to create all the
 * connection objects. The parameter string describing the target conforms to the
 * URL format as described in RFC 1808. This takes the general form: 
 * {scheme}:[{target}][{parms}] 
 * This implementation only supports connections of protocol schemes 
 * <code>btl2cap://</code> and <code>btsco://</code>,
 * and ignores all parameters. The target is defined as bluetoothAddress:UUID.
 * A valid url would be <code>btl2cap://001122334455:0001</code> or 
 * <code>btsco://00DEADBEEF</code>.
 * @author Christian Lorenz, Mauro Tortonesi
 */
public class Connector {
    public static final int READ       = 0;
    public static final int WRITE      = 0;
    public static final int READ_WRITE = 0;
    private static final String scoSchema = "btsco://";
    private static final String scoServerSchema = "btsco://localhost";
    private static final String l2capSchema = "btl2cap://";
    private static final String l2capServerSchema = "btl2cap://localhost";

    /**
     * Create and open a Connection. This Implementation only supports the 
     * <code>btl2cap://</code> protocol schema. If url begins with 
     * <code>btl2cap://localhost</code>, this method call will return a 
     * <code>javax.bluetooth.L2CAPConnectionNotifier</code>.
     * @param url The URL for the connection.
     * @return A new <code>javax.bluetooth.L2CAPConnection</code> or a 
     * <code>javax.bluetooth.L2CAPConnectionNotifier</code>
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws IOException If some other kind of I/O error occurs.
     * @see javax.bluetooth.L2CAPConnection
     * @see javax.bluetooth.L2CAPConnectionNotifier
     */
    public static Connection open(String url) throws IOException {
        try {
            /* the low-level HCI driver must be already set when the code 
             * reaches this point */
            BluetoothStack bluetooth  = BluetoothStack.getBluetoothStack();
            LocalDevice localDev      = LocalDevice.getLocalDevice();
            DiscoveryAgent discovery  = localDev.getDiscoveryAgent();
            
            Debug.println(1, "url is: " + url);

            if (url.startsWith(l2capSchema)) {
                if (url.startsWith("localhost", l2capSchema.length())) {
                    Debug.println(1, "L2CAP server connection");
                    /* we have a server connection */
                    int i = l2capServerSchema.length();
                    if (url.charAt(i++) != ':')
                        throw new IllegalArgumentException("Invalid URL, expected \':\' after " +
                                                           l2capServerSchema + ".");
                    /* ignore UUID */
                    int j = url.indexOf(';', i);
                    /* TODO: check that all the chars between i and j are digits */
                    String PSM = url.substring(i, j);
                    Short psmShort = Short.decode(PSM);
                    return new JSR82ConnectionNotifier(psmShort.shortValue());
                } else {
                    Debug.println(1, "L2CAP client connection");
                    /* retrieve address and PSM of remote device */
                    int i          = url.indexOf(':', l2capSchema.length());
                    Debug.println(1, "i is: " + i);
                    String Address = url.substring(l2capSchema.length(), i);
                    Debug.println(1, "Address is: " + Address);
                    ++i;
                    Debug.println(1, "i is: " + i);
                    int j          = url.indexOf(';', i);
                    Debug.println(1, "j is: " + j);
                    String PSM     = j > 0 ? url.substring(i, j) : url.substring(i);
                    Debug.println(1, "PSM is: " + PSM);
                    short psmShort = Short.parseShort(PSM, 16);
                    Debug.println(1, "psmShort is: " + psmShort);
                    /*
                    long AddrLong  = 0;
                    
                    try {
                        AddrLong = BluetoothAddress.JSR82ToLongRepr(Address);
                    }
                    catch (BluetoothAddressException e) {
                        throw new IOException("Invalid Bluetooth Address: " + e);
                    }
                    */
                    
                    /* TODO: change this as it is incompatible!!! */
                    RemoteDevice remoteDevice = discovery.getRemoteDevice(Address);
                    
                    if (remoteDevice == null) {
                        throw new IllegalArgumentException("Unable to locate Bluetooth Device.");
                    }

                    JSR82Channel channel = new JSR82Channel();
                    bluetooth.connectL2CAPChannel(channel, remoteDevice, psmShort);
                    return channel;
                }
            } else if (url.startsWith(scoSchema)) {
                /* notice that we don't have a colon here!!! */
                if (url.startsWith("localhost", scoSchema.length())) {
                    /* ignore all the other parameters */
                    SCOConnectionNotifier conn = null;
                    try {
                        conn = bluetooth.getSCOConnectionNotifier();
                    }
                    catch (SCOException e) {
                        System.err.println("Error getting SCOConnectionNotifier");
                        System.exit(1);
                    }
                    return conn;
                } else {
                    /* retrieve address of remote device */
                    int i          = url.indexOf(':', scoSchema.length());
                    String Address = url.substring(scoSchema.length(), i);
                    /*
                    long AddrLong  = 0;
                    
                    try {
                        AddrLong = BluetoothAddress.JSR82ToLongRepr(Address);
                    }
                    catch (BluetoothAddressException e) {
                        throw new IOException("Invalid Bluetooth Address: " + e);
                    }
                    */
                    
                    /* TODO: change this as it is incompatible!!! */
                    RemoteDevice remoteDevice = discovery.getRemoteDevice(Address);
                    
                    if (remoteDevice == null)
                        throw new IllegalArgumentException("Unable to locate Bluetooth Device.");
                    
                    SCOConnection conn = null;
                    try {
                        conn = bluetooth.getSCOConnection(remoteDevice);
                    }
                    catch (SCOException e) {
                        System.err.println("Error getting SCOConnection");
                        System.exit(1);
                    }
                    return conn;
                }
            }
        }
        catch (BluetoothStateException e) { throw new IOException("" + e); }
        catch (HCIException e) { throw new IOException("" + e); }
        throw new IllegalArgumentException("This implementation of Connector only " + 
                                           "supports btl2cap:// and btsco:// Connections.");
    }

    /**
     * Create and open a Connection.
     * @param url The URL for the connection.
     * @param mode The access mode
     * @return A new Connection object.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs. 
     * Christian Lorenz: This implementation ignores access modes.
     * This call is equivalent to Connection.open(url).
     */
    public static Connection open(String url, int mode) throws IOException {
        return open(url);
    }

    /**
     * Create and open a Connection
     * @param url The URL for the connection.
     * @param mode The access mode
     * @param timeouts A flag to indicate that the called wants timeout exceptions
     * @return A new Connection object.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs.
     * Christian Lorenz: This implementation ignores timeouts and access modes.
     * This call is equivalent to Connection.open(url).
     */
    public static Connection open(String url, int mode, boolean timeouts) throws IOException { 
        return open(url);
    }

    /**
     * Create and open a connection input stream. This method is not implemented
     * and will throw a Runtime Exception.
     * @param url The URL for the connection.
     * @return A DataInputStream.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs.
     */
    public static DataInputStream openDataInputStream(String url) { //TODO openDataInputStream
        throw new RuntimeException("This implementation of javax.microedition.io.Connector " + 
                                   "only supports open(String url).");
    }

    /**
     * Create and open a connection output stream. This method is not implemented
     * and will throw a Runtime Exception.
     * @param url The URL for the connection.
     * @return A DataOutputStream.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs.
     */
    public static DataOutputStream openDataOutputStream(String url) { //TODO openDataOutputStream
        throw new RuntimeException("This implementation of javax.microedition.io.Connector " + 
                                   "only supports open(String url).");
    }

    /**
     * Create and open a connection input stream. This method is not implemented
     * and will throw a Runtime Exception.
     * @param url The URL for the connection.
     * @return A InputStream.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs.
     */
    public static InputStream openInputStream(String url) { //TODO openInputStream
        throw new RuntimeException("This implementation of javax.microedition.io.Connector " + 
                                   "only supports open(String url).");
    }

    /**
     * Create and open a connection input stream. This method is not implemented
     * and will throw a Runtime Exception.
     * @param url The URL for the connection.
     * @return A DataOutputStream.
     * @throws IllegalArgumentException If a parameter is invalid.
     * @throws ConnectionNotFoundException If the connection cannot be found.
     * @throws IOException If some other kind of I/O error occurs.
     */
    public static OutputStream openOutputStream(String url) { //TODO openOutputStream
        throw new RuntimeException("This implementation of javax.microedition.io.Connector " + 
                                   "only supports open(String url).");
    }
}
/*
 * vim: et ts=4 sw=4
 */

