/*
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is part of the JSR82ext stack.
 * 
 * The JSR82ext stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JSR82ext stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

package org.javabluetooth.demo;

import java.io.*;
import java.util.LinkedList;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.L2CAPConnection;
import javax.microedition.io.Connector;
import org.javabluetooth.stack.hci.BluezTransport;
import org.javabluetooth.stack.hci.HCIDriver;
import org.javabluetooth.stack.BluetoothStackLocal;
import org.javabluetooth.stack.BluetoothStack;
import org.javabluetooth.stack.sco.BluezSCOConnectionFactory;
import org.javabluetooth.stack.l2cap.ASBChannel;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.BluetoothAddressException;
import org.javabluetooth.util.Debug;


public class ACLASBtest {

    public static final void main(String[] args) throws Exception {
        short acl_psm = 0;
        short asb_psm = 0;
        FileInputStream acl_file = null;
        FileInputStream asb_file = null;
        
        /* ------------------------------------------------------------
         *                      ARGUMENT VALIDATION
         * ------------------------------------------------------------ */
        
        /*
         * syntax is:
         * java ACLASBtest file_acl BD_ADDR_acl psm_acl file_asb psm_asb
         */
        if (args.length != 5) {
            System.err.println("Error: 5 parameters expected!");
            System.exit(1);
        }

        /* file_acl */
        try {
            acl_file = new FileInputStream(args[0]);
        }
        catch (FileNotFoundException e) {
            System.err.println("Error: file " + args[0] + " not found! " + e);
            System.exit(2);
        }

        /* BD_ADDR_acl */
        boolean b1 = BluetoothAddress.isValidStandardRepr(args[1]);
        boolean b2 = BluetoothAddress.isValidJSR82Repr(args[1]);
        
        Debug.println(9, "b1 is: " + b1);
        Debug.println(9, "b2 is: " + b2);
        
        if (!b1 && !b2) {
            System.err.println("Error: invalid Bluetooth address!");
            System.exit(3);
        }
        
        String acl_addr = b2 ? args[1] : BluetoothAddress.StandardToJSR82Repr(args[1]);
        
        /* psm_acl */
        try {
            acl_psm = Short.parseShort(args[2], 10);
        }
        catch (NumberFormatException e) {
            System.err.println("Error: invalid ACL PSM value!");
            System.exit(4);
        }
        
        /* file_asb */
        try {
            asb_file = new FileInputStream(args[3]);
        }
        catch (FileNotFoundException e) {
            System.err.println("Error: file " + args[3] + " not found! " + e);
            System.exit(5);
        }
       
        /* psm_asb */
        try {
            asb_psm = Short.parseShort(args[4], 10);
        }
        catch (NumberFormatException e) {
            System.err.println("Error: invalid ASB PSM value!");
            System.exit(6);
        }
        
        /* ------------------------------------------------------------
         *                  END OF ARGUMENT VALIDATION
         * ------------------------------------------------------------ */

        /* ------------------------------------------------------------
         *                   LOW LEVEL INITIALIZATION
         * ------------------------------------------------------------ */
        /* create HCI driver */
        BluezTransport transport = new BluezTransport("hci0");
        BluezSCOConnectionFactory factory = new BluezSCOConnectionFactory();
        HCIDriver driver = new HCIDriver(transport,factory);
        /* setup the Bluetooth manager */
        BluetoothStack.init(new BluetoothStackLocal(driver));
        /* now we can use BluetoothStack */
        BluetoothStack bluetooth = BluetoothStack.getBluetoothStack();
        bluetooth.send_HCI_Write_Num_Broadcast_Retransmissions((byte)0x00);
        System.out.println("Retransmissions: " + 
                           bluetooth.send_HCI_Read_Num_Broadcast_Retransmissions());
        /* ------------------------------------------------------------
         *                END OF LOW LEVEL INITIALIZATION
         * ------------------------------------------------------------ */
        
        /* ------------------------------------------------------------
         *                     REMOTE CONNECTION SETUP
         * ------------------------------------------------------------ */
        BufferedInputStream acl_lin = new BufferedInputStream(acl_file);
        BufferedInputStream asb_lin = new BufferedInputStream(asb_file);
        
        String acl_tmp = shortToHexString(acl_psm);
        Debug.println(9, "ACL: using PSM: " + acl_psm);
        Debug.println(9, "ACL: shortToHexString returned PSM: " + acl_tmp);
        
        L2CAPConnection acl_conn = null;
        try {
            acl_conn = (L2CAPConnection)Connector.open("btl2cap://" + acl_addr + 
                                                       ":" + acl_tmp);
        }
        catch (IOException e) {
            System.err.println("Error opening connection to " + acl_addr + ": " + e);
            System.exit(7);
        }
        
        Debug.println(9, "ASB: using PSM: " + asb_psm);
        ASBChannel asb_conn = new ASBChannel(asb_psm);
        /* ------------------------------------------------------------
         *                  END OF REMOTE CONNECTION SETUP
         * ------------------------------------------------------------ */

        /* ------------------------------------------------------------
         *                  THREAD CREATION AND STARTUP
         * ------------------------------------------------------------ */
        ACLxfer acl_xfer   = new ACLxfer(acl_lin, acl_conn);
        Thread  acl_thread = new Thread(acl_xfer);
        ASBxfer asb_xfer   = new ASBxfer(asb_lin, asb_conn);
        Thread  asb_thread = new Thread(asb_xfer);

        acl_thread.setPriority(Thread.MAX_PRIORITY);
        asb_thread.setPriority(Thread.MIN_PRIORITY);
        //acl_thread.setPriority(Thread.NORM_PRIORITY + 1);

        System.err.println("waiting for user input");
        /* waiting for user input */
        try {
            System.in.read();
        }
        catch (IOException e) {
            System.err.println("Unexpected I/O failure reading from " + 
                               "standard input: " + e);
            System.exit(8);
        }
        System.err.println("user input received");

        acl_thread.start();
        asb_thread.start();
        /* ------------------------------------------------------------
         *               END OF THREAD CREATION AND STARTUP
         * ------------------------------------------------------------ */
    }

    private static String shortToHexString(short s) {
        char digits[] = { '0', '1', '2', '3', '4', '5', '6', '7',
                          '8', '9', 'A', 'B', 'C', 'D', 'E', 'F' };
        char value[] = new char[4];

        value[3] = digits[((s & 0x000f))];
        value[2] = digits[((s & 0x00f0) >>  4)];
        value[1] = digits[((s & 0x0f00) >>  8)];
        value[0] = digits[((s & 0xf000) >> 12)];
        
        return new String(value);
    }
}
/*
 * vim: et ts=4 sw=4
 */

