/*
 * (c) Copyright 2003 Christian Lorenz - ALL RIGHTS RESERVED
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is part of the JavaBluetooth Stack.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

package org.javabluetooth.demo;

import java.io.*;
import java.util.LinkedList;
import javax.bluetooth.BluetoothStateException;
import javax.bluetooth.DeviceClass;
import javax.bluetooth.DiscoveryAgent;
import javax.bluetooth.DiscoveryListener;
import javax.bluetooth.LocalDevice;
import javax.bluetooth.RemoteDevice;
import javax.bluetooth.ServiceRecord;
import javax.microedition.io.Connector;
import org.javabluetooth.stack.hci.BluezTransport;
import org.javabluetooth.stack.hci.HCIDriver;
import org.javabluetooth.stack.BluetoothStackLocal;
import org.javabluetooth.stack.BluetoothStack;
import org.javabluetooth.stack.sco.BluezSCOConnectionFactory;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.BluetoothAddressException;
import org.javabluetooth.util.Debug;

/** @author Christian Lorenz */
public class SimpleInquiry implements DiscoveryListener {
    private static int INQUIRY_TIMEOUT = 60000;
    private LinkedList remoteDevices;
    private Object mutex;

    public static final void main(String[] args) throws Exception {
        if (args.length != 0) {
            System.err.println("Error: no parameters expected!");
            System.exit(1);
        }

        /* ------------------------------------------------------------
         *                   LOW LEVEL INITIALIZATION
         * ------------------------------------------------------------ */
        /* create HCI driver */
        BluezTransport transport = new BluezTransport("hci0");
        BluezSCOConnectionFactory factory = new BluezSCOConnectionFactory();
        HCIDriver driver = new HCIDriver(transport,factory);
        /* setup the Bluetooth manager */
        BluetoothStack.init(new BluetoothStackLocal(driver));
        /* now we can use BluetoothStack */
        BluetoothStack bluetooth = BluetoothStack.getBluetoothStack();
        /* ------------------------------------------------------------
         *                END OF LOW LEVEL INITIALIZATION
         * ------------------------------------------------------------ */
        
        /* create the SimpleInquiry instance */
        SimpleInquiry iq = new SimpleInquiry();
        
        iq.startInquiry();
        
        System.out.println(iq.foundDevices() + " devices found.");
        System.exit(0);
    }

    public SimpleInquiry() {
        remoteDevices = new LinkedList();
        mutex = new Object();
    }
    
    public void startInquiry() throws BluetoothStateException {
        LocalDevice localDev = LocalDevice.getLocalDevice();
        localDev.setDiscoverable(DiscoveryAgent.GIAC);
        DiscoveryAgent discovery = localDev.getDiscoveryAgent();
        discovery.startInquiry(DiscoveryAgent.GIAC, this);
        
        //Debug.println(1, "****************** Inquiry Request sent");
        try {
            //this.wait(INQUIRY_TIMEOUT);
            mutex.wait();
        }
        catch (InterruptedException e) {
            /* do nothing */
        }
    }

    /** @see javax.bluetooth.DiscoveryListener#deviceDiscovered */
    public synchronized void deviceDiscovered(RemoteDevice btDevice, DeviceClass cod) {
        System.out.println("Device Discovered: " + btDevice.getBluetoothAddress());
        String address = BluetoothAddress.JSR82ToStandardRepr(btDevice.getBluetoothAddress());
        System.out.println("Device Discovered: " + address);
        synchronized (remoteDevices) {
            remoteDevices.addLast(btDevice);
        }
        
        /*
       	try {
        	System.out.println("Device Name: " + btDevice.getFriendlyName(false));
        } catch (IOException e) {
        	e.printStackTrace();
        } 
        */
    }

    /** @see javax.bluetooth.DiscoveryListener#servicesDiscovered */
    public void servicesDiscovered(int transID, ServiceRecord[] servRecord) {
        System.out.println("Services Discovered; transID: " + transID);
    }

    /** @see javax.bluetooth.DiscoveryListener#serviceSearchCompleted */
    public void serviceSearchCompleted(int transID, int respCode) {
        System.out.println("Services Search Complete; transID: " + transID + ", respCode: " + respCode);
    }

    /** @see javax.bluetooth.DiscoveryListener#inquiryCompleted */
    public synchronized void inquiryCompleted(int discType) { 
	    System.out.println("Inquiry Complete; discType: " + discType);
        mutex.notify();
    }

    public int foundDevices() {
        int size;
        
        synchronized (remoteDevices) {
            size = remoteDevices.size();
        }

        return size;
    }
}
/*
 * vim: et ts=4 sw=4
 */

