/*
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED.
 * 
 * This file is a driver for the JavaBluetooth Stack that interfaces
 * JavaBluetooth with Bluez.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * Created on Feb 8, 2004
 * by Mauro Tortonesi
 *
 */
package org.javabluetooth.stack.hci;

import java.io.*;
import java.util.*;
import org.javabluetooth.util.Debug;
import org.javabluetooth.util.PacketQueue;
import org.javabluetooth.stack.hci.HCIException;

// TODO: implement HCITransportListener hcitl as a singleton
// synchronize with have_hcitl, not with hcitl...
/** 
 * HCITransport implementation using the Bluez protocol stack.
 * @see org.javabluetooth.stack.hci.HCITransport
 * @author Mauro Tortonesi
 */
public class BluezTransport implements HCITransport, Runnable {
    private PacketQueue tx_packet_queue;
    private HCITransportListener hcitl;
    private boolean isStarted;

    static {
        /* Load libBluezTransport.so (which must be in one of the directories
         * in /etc/ld.so.conf or in the LD_LIBRARY_PATH environment variable). */
        System.loadLibrary("BluezTransport");
    }

    private native int startBluez(PacketQueue tx_queue, HCITransportListener hcitl);
    
    /**
     * BluezTransport constructor.
     */
    public BluezTransport(String InterfaceName) {
        /* For the moment, the InterfaceName parameter is ignored. In future, 
         * developers will be able to choose the Bluez interface, e.g., "hciX",
         * they wish to use. */
        tx_packet_queue = new PacketQueue();
        isStarted = false;
        hcitl = null;
        // reset();
    }
    
    public void run() {
        /* we can't use a synchronized(hcitl) { } construct here, since 
         * we would allow access to hcitl only to this thread.
         * let's hope that there won't be any race condition :-( */
        if (hcitl != null && tx_packet_queue != null) {
            //Debug.println(11, "run() like hell ;-)");
            int ret = startBluez(tx_packet_queue, hcitl);
            //Debug.println(11, "out of run(), startBluez returned: " + ret);
        } 
    }
    
    /**
     * Queues a packet for transmission to the HCI Host Controller. 
     * The bytes are send via the Bluez interface to the Bluetooth Host Controller.
     * @see org.javabluetooth.stack.hci.HCITransport#sendPacket(byte[])
     * @throws HCIException 
     */
    public synchronized void sendPacket(byte[] packet) throws HCIException {
        //Debug.println(1, "HCI: Queuing Packet for TX: ", packet);
        try {
            tx_packet_queue.insert(packet);
        } 
        catch (Exception e) {
            throw new HCIException("Error adding packet to transmission queue:" +
                                   e.toString());
            
        }
        //Debug.println(1, "HCI: Packet queued.");
    }

    /** 
     * Resets the Bluetooth Host Controller.
     * @see org.javabluetooth.stack.hci.HCITransport#reset() 
     * @throws HCIException */
    public synchronized void reset() throws HCIException {
        /* yet to be implemented... */
    }
    
    /**
     * Registers a HCITransportListener interface.
     * Only one HCITransportListener interface can be registered.
     * @throws HCIException
     */
    public synchronized void registerHCITransportListener(HCITransportListener _hcitl)
                                                          throws HCIException {
        //Debug.println(1, "HCI: registering a new HCITransportListener.");
        
        if (_hcitl == null)
            throw new HCIException("Invalid argument.");
        
        if (hcitl != null) {
            throw new HCIException("Cannot register more than one HCITransportListener.");
        } else {
            hcitl = _hcitl;
        }

        Thread t = new Thread(this);
        t.start();
    }
}
/* 
 * vim: et ts=4 sw=4 
 */

