/*
 * (c) Copyright 2003 Christian Lorenz - ALL RIGHTS RESERVED
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is part of the JSR82ext stack.
 * 
 * The JSR82ext stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JSR82ext stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * Created on May 21, 2003
 * by Christian Lorenz
 */

package org.javabluetooth.stack.hci;

import java.util.*;
import javax.bluetooth.RemoteDevice;
import org.javabluetooth.stack.l2cap.L2CAPLink;
import org.javabluetooth.stack.sco.SCOConnection;
import org.javabluetooth.stack.sco.SCOConnectionFactory;
import org.javabluetooth.stack.sco.SCOConnectionNotifier;
import org.javabluetooth.stack.sco.SCOException;
import org.javabluetooth.util.Debug;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.Semaphore;

/** 
 * The HCIDriver class implements the upper (and hardware-independent) layer 
 * of a JavaBluetooth driver. The hardware-specific part of a JavaBluetooth 
 * driver is implemented by HCITransport-derived classes.
 * @see org.javabluetooth.stack.hci.HCITransport
 * @author Christian Lorenz, Mauro Tortonesi
 */
public class HCIDriver implements HCITransportListener {
    /* packet type codes */
    private static final byte PACKET_TYPE_COMMAND                    = 0x01;
    private static final byte PACKET_TYPE_ACL                        = 0x02;
    private static final byte PACKET_TYPE_SCO                        = 0x03;
    private static final byte PACKET_TYPE_EVENT                      = 0x04;
    /* mauro: java sucks badly! how can a language that defines a "byte" data
     * type as a ***__SIGNED__*** (!!!) 8 bits word be considered serious? 
     * come on, this is ridiculous... */
    private static final byte PACKET_TYPE_UNKNOWN                    = -128; 
    /* these values consider also the first byte that contains the packet
     * type */
    private static final byte COMMAND_PACKET_HEADER_SIZE             = 0x04;
    private static final byte ACL_PACKET_HEADER_SIZE                 = 0x05;
    private static final byte SCO_PACKET_HEADER_SIZE                 = 0x04;
    private static final byte EVENT_PACKET_HEADER_SIZE               = 0x03;
    private static final byte MAX_PACKET_HEADER_SIZE                 = 0x05;
    /* HCI event codes */
    private static final byte HCI_EVENT_INQUIRY_COMPLETE             = 0x01;
    private static final byte HCI_EVENT_INQUIRY_RESULT               = 0x02;
    private static final byte HCI_EVENT_CONNECTION_COMPLETE          = 0x03;
    private static final byte HCI_EVENT_DISCONNECTION_COMPLETE       = 0x05;
    private static final byte HCI_EVENT_REMOTE_NAME_REQUEST_COMPLETE = 0x07;
    private static final byte HCI_EVENT_COMMAND_COMPLETE             = 0x0E;
    private static final byte HCI_EVENT_COMMAND_STATUS               = 0x0F;
    private static final byte HCI_EVENT_NUMBER_OF_COMPLETED_PACKETS  = 0x13;
    
    /* data members */
    private HCITransport hciTransport;
    private SCOConnectionFactory scoConnectionFactory;
    private byte[] commandResponse;
    private short commandResponseOpCode = 0;
    private Hashtable connectionHandles = new Hashtable();
    private Hashtable remoteAddresses   = new Hashtable();
    /* mauro: maybe we should change the type of hciReceivers from Vector 
     * to LinkedList? */
    private Vector hciReceivers         = new Vector();
    private byte[] headerBuffer         = new byte[MAX_PACKET_HEADER_SIZE];
    private int headerBufferIndex       = 0;
    private int headerLength            = 0;
    private byte[] packetBuffer;
    private int packetBufferIndex;
    private Object sendCommandMutex;
    private Object commandResponseMutex;
    private Semaphore waitForCommandResponseMutex;
    private short acl_packet_len;
    private short sco_packet_len;
    private short acl_buffers_num;
    private short sco_buffers_num;
    private Semaphore acl_buffers;

    /* mauro: we don't declare this stuff as static because we want to be able to use 
     * several HCIDrivers with different HCITransports and SCOConnectionFactories
     * at the same time. */
    public HCIDriver(HCITransport _hciTransport, 
                     SCOConnectionFactory _scoConnectionFactory)
                     throws HCIException {
        hciTransport = _hciTransport;
        hciTransport.registerHCITransportListener(this);
        scoConnectionFactory = _scoConnectionFactory;
        sendCommandMutex = new Object();
        commandResponseMutex = new Object();
        waitForCommandResponseMutex = new Semaphore();

        /* we first set this to null because we don't know the number of 
         * controller buffers yet */
        acl_buffers = null;

        /* send "Read Buffer Size" command (ogf = 0x04, ocf = 0x0005) and 
         * get information on the number and size of the controller buffers */
        byte[] data = { 0x01, 0x05, (0x04 << 2), 0x00 }; 
        byte[] resultData = null;
        try {
            resultData = send_HCI_Command_Packet(data);
        }
        catch (HCIException e) {
            System.out.println("Fatal error with Read Buffer Size command!");
            System.exit(1);
        }
        /* read returned data */
        //Debug.println(9, "resultData.length is: " + resultData.length);
        byte code = resultData[1];
        //Debug.println(9, "code is: " + code);
        byte len = resultData[2];
        //Debug.println(9, "len is: " + len);
        byte status = resultData[6];
        //Debug.println(9, "status is: " + status);
        if (resultData.length != 0x0E || code != 0x0E || len != 0x0B || status != 0) 
            throw new RuntimeException("Error with Read Buffer Size command!");
        acl_packet_len  = (short)(((short)resultData[ 7] & 0xff) |
                                  ((short)resultData[ 8] & 0xff) << 8);
        sco_packet_len  = (short)resultData[ 9];
        acl_buffers_num = (short)(((short)resultData[10] & 0xff) |
                                  ((short)resultData[11] & 0xff) << 8);
        sco_buffers_num = (short)(((short)resultData[12] & 0xff) |
                                  ((short)resultData[13] & 0xff) << 8);
        Debug.println(8, "acl packet length is: " + acl_packet_len);
        Debug.println(8, "sco packet length is: " + sco_packet_len);
        Debug.println(8, "the number of acl packet buffer is: " + acl_buffers_num);
        Debug.println(8, "the number of sco packet buffer is: " + sco_buffers_num);

        //acl_buffers = new Semaphore(acl_buffers_num, true);
        acl_buffers = new Semaphore(acl_buffers_num);
    }

    public short get_ACL_MTU() {
        return acl_packet_len;
    }
    
    private void sendPacket(byte[] packet) throws HCIException {
        // with the semaphore construct, we acquire the lock on one of the
        // hci device internal buffers
        try {
            acl_buffers.acquire();
        }
        catch (InterruptedException e) { }
        hciTransport.sendPacket(packet);
    }
    
    /* use synchronized or synchronized(hciReceivers) {} ? */
    public void registerHCIReceiver(HCIReceiver receiver) {
        if (!hciReceivers.contains(receiver)) 
            hciReceivers.addElement(receiver);
    }

    /* use synchronized or synchronized(hciReceivers) {} ? */
    public void unregisterHCIReceiver(HCIReceiver receiver) { 
        hciReceivers.removeElement(receiver); 
    }

    /* use synchronized? */
    public void registerL2CAPLink(L2CAPLink link) {
        Short handle = new Short(link.connectionHandle);
        connectionHandles.put(handle, link);
        Long remoteAddress = new Long(link.remoteAddress);
        remoteAddresses.put(remoteAddress, link);
        Debug.println(11, "HCIDriver: registering L2CAPLink for address: " + 
                      BluetoothAddress.LongToStandardRepr(link.remoteAddress) +
                      ", handle: " + link.connectionHandle);
    }

    /* use synchronized? */
    public void unregisterL2CAPLink(L2CAPLink link) {
        connectionHandles.remove(new Short(link.connectionHandle));
        remoteAddresses.remove(new Long(link.remoteAddress));
    }

    public L2CAPLink getL2CAPLink(long remoteAddress) throws HCIException {
        Long remoteAddressLong = new Long(remoteAddress);
        L2CAPLink link = (L2CAPLink)remoteAddresses.get(remoteAddressLong);
        if (link == null) {
            /* defaults are:
             * packet type = 0x8000 (DH5 packets) 
             * pageScanRepMode = 0x02 (don't know why, taken from bluez)
             * pageScanMode = 0x00
             * clockOffset = 0x0000
             * roleSwitch = 0x01 (allow role switch)
             */            
            byte result = send_HCI_LC_Create_Connection(remoteAddress, (short)0x8000, 
                                                        (byte)0x02, (byte)0x00, 
                                                        (short)0x0000, (byte)0x01);
            if (result != 0) 
                throw new HCIException("Create Connection failed (" + result + ").");
            short timeout = 0;
            while (link == null) {
                try {
                    Thread.sleep(1000);
                    timeout++;
                }
                catch (InterruptedException e) {
                    /* do nothing */
                }
                if (timeout == 50) /* timeout is 50 seconds */
                    throw new HCIException("Create Connection timed out.");
                link = (L2CAPLink)remoteAddresses.get(remoteAddressLong);
            }
        }
        return link;
    }
    
    public SCOConnection getSCOConnection(RemoteDevice remoteDevice) throws SCOException {
        return scoConnectionFactory.getSCOConnection(remoteDevice);
    }

    public SCOConnectionNotifier getSCOConnectionNotifier() throws SCOException {
        return scoConnectionFactory.getSCOConnectionNotifier();
    }

    /* this function provides a synchronous interface to send HCI command
     * packets */
    public byte[] send_HCI_Command_Packet(byte[] cmdPacket) throws HCIException {
        short opCode = (short)((cmdPacket[2] << 8) | 
                               (cmdPacket[1] & 0xff));
        byte[] result;
        
        Debug.println(11, "HCI: Sending Command: " + opCode);
        
        /* we cannot send more than one command together */
        synchronized (sendCommandMutex) {
            if (commandResponse != null) {
                Debug.println(1, "Error in send_HCI_Command_Packet: expected commandResponse == null!!!");
            }

            /* initialize stuff and send command */
            commandResponseOpCode = opCode;

            /* send the packet */
            if (acl_buffers == null) {
                hciTransport.sendPacket(cmdPacket);
            } else {
                sendPacket(cmdPacket);
            }

            /* wait until the command response arrives */
            try {
                waitForCommandResponseMutex.acquire();
            }
            catch (InterruptedException e) { }
            
            result = commandResponse;
            if (result == null) {
                Debug.println(1, "Error in send_HCI_Command_Packet: expected commandResponse != null!!!");
            }
            commandResponseOpCode = 0;
            commandResponse = null;
        }

        return result;
    }

    public synchronized void send_HCI_Data_Packet(byte[] dataPacket) throws HCIException {
            Debug.println(11, "HCI: Sending Data Packet.");
            sendPacket(dataPacket);
    }

    /**
     * This command will cause the Link Manager to create a connection
     * to the Bluetooth device with the BD_ADDR specified by the command
     * parameters. It is triggered by <code>HCITransport.getL2CAPLink()</code>.  
     * For details see Page 568 of the Bluetooth Core Specification Version 1.1
     * @param bd_addr
     * @param packetType
     * @param clockOffset
     * @param allowRoleSwitching
     * @return return 0x00 if the command succeeded. 0x01-0xFF if the command failed. 
     * See Table 6.1 on page 766 for list of Error Codes.
     * @throws HCIException
     */
    public byte send_HCI_LC_Create_Connection(long bd_addr, short packetType, byte pageScanRepMode, 
                                              byte pageScanMode, short clockOffset, 
                                              byte allowRoleSwitching) throws HCIException {
        byte[] data = { 0x01, 0x05, 0x04, 0x0d, 
                        (byte)((bd_addr) & 0xff), 
                        (byte)((bd_addr >> 8) & 0xff), 
                        (byte)((bd_addr >> 16) & 0xff), 
                        (byte)((bd_addr >> 24) & 0xff), 
                        (byte)((bd_addr >> 32) & 0xff), 
                        (byte)((bd_addr >> 40) & 0xff), 
                        (byte)((packetType) & 0xff),
                        (byte)((packetType >> 8) & 0xff), 
                        (byte)pageScanRepMode, // Page scan repetition mode 1 byte
                        (byte)pageScanMode, // Page scan mode 1 byte
                        (byte)((clockOffset) & 0xff), // Clock offset 2 bytes
                        (byte)((clockOffset >> 8) & 0xff), 
                        (byte)allowRoleSwitching
        };
        byte[] resultData = send_HCI_Command_Packet(data);
        return resultData[3];
    }

    /**
     * The Disconnection command is used to terminate an existing connection.
     * The Connection Handle indicates which connection is to be disconnected.
     * It is triggered by <code>L2CAPLink.close()</code>. When the Host 
     * Controller receives the Disconnect command, it sends the Command Status
     * event to the Host. The Disconnection Complete event will occur at each 
     * Host when the termination of the connection has completed, and indicates
     * that this command has been completed.
     * For details see Page 571 of the Bluetooth Core Specification Version 1.1
     * @param connectionHandle Connection Handle for the connection being disconnected.
     * @return 0x00 if the command succeeded. 0x01-0xFF if the command failed. 
     * See Table 6.1 on page 766 for list of Error Codes.
     * @throws HCIException
     */
    public byte send_HCI_LC_Disconnect(short connectionHandle) throws HCIException {
        byte[] data = { 0x01, 0x06, 0x04, 03, (byte)((connectionHandle) & 0xff), 
                        (byte)((connectionHandle >> 8) & 0xff), 0x13 };
        byte[] resultData = send_HCI_Command_Packet(data);
        return resultData[3];
    }

    /**
     * The Reset command will reset the Host Controller and the Link Manager. 
     * After the reset is completed, the current operational state will be lost, 
     * the Bluetooth device will enter standby mode and the Host Controller will
     * automatically revert to the default values for the parameters for which 
     * default values are defined in the specification.
     * For details see Page 622 of the Bluetooth Core Specification Version 1.1.
     * @return 0x00 if the command succeeded. 0x01-0xFF if the command failed. 
     * See Table 6.1 on page 766 for list of Error Codes.
     * @throws HCIException
     */
    public byte send_HCI_HC_Reset() throws HCIException {
            byte[] data = { 0x01, 0x03, 0x0C, 0x00 };
            byte[] resultData = send_HCI_Command_Packet(data);
            return resultData[6];
    }

    /**
     * Accepts data from the low-level HCITransport driver. This function 
     * performs processing of incoming data and forwards the packets to 
     * the registered HCIReceivers. 
     */
    public void receiveData(byte[] incomingBytes, int incomingLength) {
        /*
        Debug.println(1, "HCI: Received " + incomingLength + ":" +
                      incomingBytes.length + " bytes of data (" + 
                      Debug.printByteArray(incomingBytes) + ")");
        */
        Debug.println(11, "HCIDriver: in receiveData");
        
        int incomingBytesIndex = 0;
        
        /* process all available data */ 
        while (incomingBytesIndex < incomingLength) {
            
            /* if we have a new packet, decode packet type and find out the header length */
            if (headerBufferIndex == 0) {
                /* read packet type */
                switch (incomingBytes[0]) {
                case PACKET_TYPE_ACL:
                    headerLength = ACL_PACKET_HEADER_SIZE;
                    break;
                case PACKET_TYPE_EVENT:
                    headerLength = EVENT_PACKET_HEADER_SIZE;
                    break;
                case PACKET_TYPE_SCO:
                    /* SCO packets coming from the HCI are read and then silently discarded */
                    headerLength = SCO_PACKET_HEADER_SIZE;
                    break;
                case PACKET_TYPE_COMMAND:
                    /* we don't expect command packets coming from the HCI - is this an error? */
                    headerLength = COMMAND_PACKET_HEADER_SIZE;
                    break;
                case PACKET_TYPE_UNKNOWN:
                    /* we don't expect unknown packets coming from the HCI - this is an error */
                default:
                    System.err.println("HCI: Received Invalid Packet Header: " + 
                                       Debug.printByteArray(incomingBytes));
                    /* perhaps we should try to recover from this error and keep going on? */
                    System.exit(1);
                }
            }
            
            if (headerBufferIndex < headerLength) {
                /* header is still incomplete */
                int length = headerLength - headerBufferIndex;
                if (length > incomingLength - incomingBytesIndex)
                    length = incomingLength - incomingBytesIndex;
                System.arraycopy(incomingBytes, incomingBytesIndex, headerBuffer,
                                 headerBufferIndex, length);
                incomingBytesIndex += length;
                headerBufferIndex += length;
                
                /* if header is complete, process header */
                if (headerBufferIndex == headerLength) {
                    /* create packet and copy header into packet */
                    switch (headerBuffer[0]) {
                        case PACKET_TYPE_COMMAND: /* read the packet and ignore it */
                            /* create a new packet buffer to read incoming data into */
                            packetBuffer = new byte[COMMAND_PACKET_HEADER_SIZE + 
                                                    (short)((short)headerBuffer[3] & 0xff)];
                            /* copy the header into the packet */
                            System.arraycopy(headerBuffer, 0, packetBuffer, 0, headerLength);
                            /* the rest of the packet must be copied into packetBuffer 
                             * after the header, starting from the headerLength offset */
                            packetBufferIndex = headerLength;
                            break;
                        case PACKET_TYPE_ACL:
                            /* create a new packet buffer to read incoming data into */
                            packetBuffer = new byte[ACL_PACKET_HEADER_SIZE + 
                                                    (short)(((short)headerBuffer[4] & 0xff) << 8 | 
                                                            ((short)headerBuffer[3] & 0xff))];
                            /* copy the header into the packet */
                            System.arraycopy(headerBuffer, 0, packetBuffer, 0, headerLength);
                            /* the rest of the packet must be copied into packetBuffer 
                             * after the header, starting from the headerLength offset */
                            packetBufferIndex = headerLength;
                            break;
                        case PACKET_TYPE_SCO: /* read the packet and ignore it */
                            /* create a new packet buffer to read incoming data into */
                            packetBuffer = new byte[SCO_PACKET_HEADER_SIZE + 
                                                    (short)((short)headerBuffer[3] & 0xff)];
                            /* copy the header into the packet */
                            System.arraycopy(headerBuffer, 0, packetBuffer, 0, headerLength);
                            /* the rest of the packet must be copied into packetBuffer 
                             * after the header, starting from the headerLength offset */
                            packetBufferIndex = headerLength;
                            break;
                        case PACKET_TYPE_EVENT:
                            /* create a new packet buffer to read incoming data into */
                            packetBuffer = new byte[EVENT_PACKET_HEADER_SIZE + 
                                                    (short)((short)headerBuffer[2] & 0xff)];
                            /* copy the header into the packet */
                            System.arraycopy(headerBuffer, 0, packetBuffer, 0, headerLength);
                            /* the rest of the packet must be copied into packetBuffer 
                             * after the header, starting from the headerLength offset */
                            packetBufferIndex = headerLength;
                            break;
                        //case PACKET_TYPE_UNKNOWN:
                            /* we don't expect unknown packets coming from the HCI - this is an error */
                        default:
                            System.err.println("HCI: Received Invalid Packet Header: " + 
                                               Debug.printByteArray(incomingBytes));
                            /* perhaps we should try to recover from this error and keep going on? */
                            System.exit(1);
                    }
                }
            } else { 
                /* header is complete, copy data to packet */
                int length = packetBuffer.length - packetBufferIndex;
                if (length > incomingLength - incomingBytesIndex) 
                    length = incomingLength - incomingBytesIndex;
                System.arraycopy(incomingBytes, incomingBytesIndex, packetBuffer, 
                                 packetBufferIndex, length);
                incomingBytesIndex += length;
                packetBufferIndex += length;
                
                /* packet is complete */
                if (packetBufferIndex == packetBuffer.length) { 
                    /* this will lead to a new packet creation */
                    headerBufferIndex = 0; 
                    dispatchPacket(packetBuffer);
                }
            }
        }
        
        Debug.println(11, "HCIDriver: out of receiveData");
    }

    private void dispatchPacket(byte[] hciPacket) {
        Debug.println(11, "HCIDriver: in dispatchPacket" /*, hciPacket */);
        switch (hciPacket[0]) // Packet Type
        {
            case PACKET_TYPE_EVENT:
                switch (hciPacket[1]) // Event Type
                {
                    case HCI_EVENT_INQUIRY_COMPLETE:
                        Debug.println(11, "Received Event Inquiry Complete");
                        receive_HCI_Event_Inquiry_Complete(hciPacket);
                        break;
                    case HCI_EVENT_INQUIRY_RESULT:
                        Debug.println(11, "Received Event Inquiry Result");
                        receive_HCI_Event_Inquiry_Result(hciPacket);
                        break;
                    case HCI_EVENT_COMMAND_COMPLETE:
                        Debug.println(11, "Received Event Command Complete");
                        receive_HCI_Event_Command_Complete(hciPacket);
                        break;
                    case HCI_EVENT_COMMAND_STATUS:
                        Debug.println(11, "Received Event Command Status");
                        receive_HCI_Event_Command_Status(hciPacket);
                        break;
                    case HCI_EVENT_CONNECTION_COMPLETE:
                        Debug.println(11, "Received Event Connection Complete");
                        receive_HCI_Event_Connection_Complete(hciPacket);
                        break;
                    case HCI_EVENT_DISCONNECTION_COMPLETE:
                        Debug.println(11, "Received Event Disconnection Complete");
                        receive_HCI_Event_Disconnection_Complete(hciPacket);
                        break;
                    case HCI_EVENT_REMOTE_NAME_REQUEST_COMPLETE:
                        Debug.println(11, "Received Event Name Request Complete");
                        receive_HCI_Event_Remote_Name_Request_Complete(hciPacket);
                        break;
                    case HCI_EVENT_NUMBER_OF_COMPLETED_PACKETS:
                        Debug.println(11, "Received Event Number Of Completed Packets");
                        receive_HCI_Event_Number_Of_Completed_Packets(hciPacket);
                        break;
                    default:
                        System.err.println("HCI: Received Unknown Event Packet: " + 
                                           Debug.printByteArray(hciPacket));
                }
                break;
            case PACKET_TYPE_ACL:
                //Debug.println(9, "HCI: Received Data Packet:", hciPacket);
                Debug.println(11, "HCI: Received Data Packet");
                receive_HCI_Data_Packet(hciPacket);
                break;
            default:
                System.err.println("HCI: Received Packet of Unknown Type: " +
                                   Debug.printByteArray(hciPacket));
        }
        Debug.println(11, "HCIDriver: out from dispatchPacket" /*, hciPacket */);
    }

    private void receive_HCI_Event_Inquiry_Complete(byte[] eventPacket) {
        //Debug.println(1, "HCI: Received Inquiry Complete Event.");
        Enumeration receivers = hciReceivers.elements();
        while (receivers.hasMoreElements()) {
            HCIReceiver receiver = (HCIReceiver)receivers.nextElement();
            receiver.receive_HCI_Event_Inquiry_Complete(eventPacket);
        }
    }

    private void receive_HCI_Event_Inquiry_Result(byte[] eventPacket) {
        //Debug.println(1, "HCI: Received Inquiry Result Event.");
        Enumeration receivers = hciReceivers.elements();
        while (receivers.hasMoreElements()) {
            HCIReceiver receiver = (HCIReceiver)receivers.nextElement();
            receiver.receive_HCI_Event_Inquiry_Result(eventPacket);
        }
    }

    private void receive_HCI_Event_Connection_Complete(byte[] eventPacket) {
        if (eventPacket[3] == 0) { //is connected
            L2CAPLink conn = new L2CAPLink(this, eventPacket);
            Debug.println(11, "HCI: Received Connection Complete Event: " + 
                          BluetoothAddress.LongToStandardRepr(conn.remoteAddress));
        }
        else Debug.println(1, "HCI: Received Connection Complete Event: Create Connection failed.");
    }

    private void receive_HCI_Event_Disconnection_Complete(byte[] eventPacket) {
        short connectionHandle = (short)(((short)eventPacket[5] & 0x0f) << 8 | 
                                         ((short)eventPacket[4] & 0xff));
        L2CAPLink link = (L2CAPLink)connectionHandles.remove(new Short(connectionHandle));
        if (link != null) {
            Debug.println(11, "HCI: Received Disconnection Complete Event: " + link.remoteAddress);
            link.wasDisconnected();
        }
        else Debug.println(1, "HCI: Received unexpected Disconnection Complete Event.");
    }

    private void receive_HCI_Event_Remote_Name_Request_Complete(byte[] eventPacket) {
        //Debug.println(1, "HCI: Remote Name Request Complete Event.");
        Enumeration receivers = hciReceivers.elements();
        while (receivers.hasMoreElements()) {
            HCIReceiver receiver = (HCIReceiver)receivers.nextElement();
            receiver.receive_HCI_Event_Remote_Name_Request_Complete(eventPacket);
        }
    }

    private void receive_HCI_Event_Command_Complete(byte[] packetData) {
        //TODO catch Reset Event and reset rest of stack
        short opCode = (short)((packetData[5] << 8) | (packetData[4] & 0xff));
        //Debug.println(1, "HCI: Received Command Complete Event: " + opCode);
        /* wait until the command response arrives */
        synchronized (commandResponseMutex) {
            if (opCode == commandResponseOpCode) {
                Debug.println(11, "set commandResponse = ", packetData);
            } else {
                Debug.println(1, "Error, expected opcode " + commandResponseOpCode + "!");
            }
            commandResponse = packetData;
            waitForCommandResponseMutex.release();
        }
    }

    private void receive_HCI_Event_Command_Status(byte[] packetData) {
        short opCode = (short)((packetData[6] << 8) | (packetData[5] & 0xff));
        //Debug.println(1, "HCI: Received Command Status Event: " + opCode);
        /* wait until the command response arrives */
        synchronized (commandResponseMutex) {
            if (opCode == commandResponseOpCode) {
                Debug.println(11, "set commandResponse = ", packetData);
            } else {
                Debug.println(1, "Error, expected opcode " + commandResponseOpCode + "!");
            }
            commandResponse = packetData;
            waitForCommandResponseMutex.release();
        }
    }

    private void receive_HCI_Event_Number_Of_Completed_Packets(byte[] packetData) {
        int len = packetData[2];        
        // number of connection handles
        int conn_num = packetData[3];

        /*
        Debug.println(1, "HCIDriver: in receive_HCI_Event_Number_Of_Completed_Packets" +
                      " len is: " + len + " conn_num is: " + conn_num + 
                      " packetData.length is: " + packetData.length);
        */
        if (len + EVENT_PACKET_HEADER_SIZE != packetData.length || len != 1 + conn_num * 4) {
            System.err.println("Error, malformed Number Of Completed Packets Event packet!");
            System.exit(1);
        }
        
        for (int i = 0, offset = 4; i < conn_num; ++i, offset += 4) {
            short handle  = (short)(((short)packetData[offset  ] & 0xff) |
                                    ((short)packetData[offset+1] & 0xff) << 8);
            short packets = (short)(((short)packetData[offset+2] & 0xff) |
                                    ((short)packetData[offset+3] & 0xff) << 8);
                                   
            // check that this is a valid/registered acl connection
            if (handle != L2CAPLink.ASBConnectionHandle && 
                !connectionHandles.containsKey(new Short(handle))) {
                System.err.println("Error, unknown connection handle in " + 
                                   "Number Of Completed Packets Event packet!");
                // keep on going for the moment...
                continue;                
            } 

            for (int j = 0; j < packets; ++j)
                acl_buffers.release();
            
            // TODO: add support also for sco connections
        }

    }
    
    private void receive_HCI_Data_Packet(byte[] data) {
        short handle = (short)(((short)data[2] & 0x0f) << 8 | 
                               ((short)data[1] & 0xff));
        L2CAPLink link = (L2CAPLink)connectionHandles.get(new Short(handle));
        Debug.println(11, "HCI: Received Data Packet on handle " + handle);
        if (link != null) {
            link.receiveData(data);
            Debug.println(11, "HCI: Received Data Packet from " + link.remoteAddress);
        }
        else System.err.println("HCI: Unable to deliver Data Packet. No open link with handle " + 
                                handle + ". : " + Debug.printByteArray(data));
    }

    public static short packOpcode(short ogf, short ocf) {
        // TODO: 
        // assert ogf <= 0x3d 
        // (0x3e reserved for log testing, 
        //  0x3f reserved for vendor specific debug commands)
        // assert ocf <= 0x03ff
        return (short)((ocf & 0x03ff) | (ogf << 10));
    }
    
    public static short ogfFromOpcode(short opcode) {
        // TODO: 
        // assert ogf <= 0x3d
        short ogf = (short)(opcode >> 10);
        return ogf;
    }
    
    public static short ocfFromOpcode(short opcode) {
        short ocf = (short)(opcode & 0x03ff);
        return ocf; 
    }
    
    public static String opcodeToString(short opcode) {
        return "ogf = " + ogfFromOpcode(opcode) + ", ocf = " + ocfFromOpcode(opcode);
    }
}
/*
 * vim: et ts=4 sw=4
 */

