/*
 * (c) Copyright 2003 Christian Lorenz - ALL RIGHTS RESERVED
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is part of the JSR82ext stack.
 * 
 * The JSR82ext stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JSR82ext stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * Created on Jun 22, 2003
 * by Christian Lorenz
 */

package org.javabluetooth.stack.l2cap;

import java.io.*;
import org.javabluetooth.stack.BluetoothStack;
import org.javabluetooth.stack.hci.HCIDriver;
import org.javabluetooth.stack.hci.HCIException;
//import org.javabluetooth.stack.sdp.SDPServer;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.BluetoothAddressException;
import org.javabluetooth.util.Debug;
import java.util.Hashtable;

/** 
 * This class represents a link between two Bluetooth Devices. 
 * It parses data packets into their appropriate L2CAP Channels, and manages
 * the Control Channel for each link.
 * @author Christian Lorenz
 */
public class L2CAPLink implements L2CAPSender {
    private static final byte L2CAP_PACKET_BOUNDRY_FLAG_FIRST      = 0x02;
    private static final byte L2CAP_PACKET_BOUNDRY_FLAG_CONTINUING = 0x01;
    private static final byte L2CAP_COMMAND_REJECT                 = 0x01;
    private static final byte L2CAP_CONNECTION_REQUEST             = 0x02;
    private static final byte L2CAP_CONNECTION_RESPONSE            = 0x03;
    private static final byte L2CAP_CONFIGURE_REQUEST              = 0x04;
    private static final byte L2CAP_CONFIGURE_RESPONSE             = 0x05;
    private static final byte L2CAP_DISCONNECTION_REQUEST          = 0x06;
    private static final byte L2CAP_DISCONNECTION_RESPONSE         = 0x07;
    private static final byte L2CAP_ECHO_REQUEST                   = 0x08;
    private static final byte L2CAP_ECHO_RESPONSE                  = 0x09;
    private static final byte L2CAP_INFORMATION_REQUEST            = 0x0A;
    private static final byte L2CAP_INFORMATION_RESPONSE           = 0x0B;
    private static final byte L2CAP_LINK_CLOSED                    = 0x00;
    private static final byte L2CAP_LINK_OPEN                      = 0x01;
    /* number of channels per link supported by this implementation */
    private static final int NUM_OF_CHANNELS                       = 16;
    /* the first NUM_OF_RESERVED_CHANNELS channels are reserved and cannot
     * be assigned to L2CAP connections:
     *  - channel  0x0000: null identifier - invalid channel
     *  - channel  0x0001: signalling channel
     *  - channel  0x0002: active slave broadcast
     *  - channels 0x0003 - 0x003F: reserved
     *  - channels 0x0040 - 0xFFFF: dynamically allocated
     */
    private static final int NUM_OF_RESERVED_CHANNELS              = 0x0040;
    private static final short NULL_CHANNEL                        = 0x00;
    private static final short SIGNALLING_CHANNEL                  = 0x01;
    private static final short BROADCAST_CHANNEL                   = 0x02;
    /* asb channels */
    private static final int NUM_OF_ASB_CHANNELS                   = 16;
    private static final int START_OF_ASB_CHANNELS                 = 0x1000;
    /* packet boundary flags */
    private static final byte CONTINUING_PACKET                    = 0x01;
    private static final byte FIRST_PACKET                         = 0x02;
    /* broadcast flags, host to controller */
    private static final byte H2C_NO_BROADCAST                     = 0x00;
    private static final byte H2C_ACTIVE_SLAVE_BROADCAST           = 0x01;
    private static final byte H2C_PARKED_SLAVE_BROADCAST           = 0x02;
    /* broadcast flags, controller to host */
    private static final byte C2H_NO_BROADCAST                     = 0x00;
    private static final byte C2H_RECEIVED_FROM_NOT_PARKED_SLAVE   = 0x01;
    private static final byte C2H_RECEIVED_FROM_PARKED_SLAVE       = 0x02;
    // this is the size of L2CAP header for normal connection-oriented packets
    // the header of broadcast packet have also a variable size psm field
    private static final short L2CAP_HDR_LEN                       = 4;
    // 1 byte for packet type and 4 bytes for the real header
    private static final short HCI_ACL_HDR_LEN                     = 5;
    // the bluetooth specifications state that size of psm field should be 
    // variable, but in the real world everyone uses two bytes
    private static final short PSM_FIELD_SIZE                      = 2;
    /* this should be public, as incoming connection requests should not use
     * the same handle */
    public static final short ASBConnectionHandle                  = 0x0EFE;
    private static ASBChannel asbChannels[]                        = null;
    private static Hashtable asbChannelsByPSM                      = null;
    private static Object asbChannelsLock                          = null;
    private static int bcast_packet_sent                           = 0;
    
    private HCIDriver hciDriver;
    public short connectionHandle;
    public long remoteAddress;
    public String remoteAddressString;
    private byte linkState;
    //private SDPServer sddb;
    private L2CAPChannel[] channels;
    private byte[] packetBuffer;
    private int packetBufferIndex;
    private short packetBufferChannelID;
    private byte identifierCount;
    private short linkMTU;

    static {
        asbChannels = new ASBChannel[NUM_OF_ASB_CHANNELS];
        asbChannelsByPSM = new Hashtable();
        asbChannelsLock = new Object();
        bcast_packet_sent = 0;
    }
    
    /**
     * @param hciDriver The HCIDriver associated with this link.
     * @param data The Connection Complete HCI Event Packet, which triggered the creation of this object.
     */
    public L2CAPLink(HCIDriver hciDriver, byte[] data) {
        this.hciDriver = hciDriver;
        //this.sddb = SDPServer.getSDPServer();
        channels = new L2CAPChannel[NUM_OF_CHANNELS];
        connectionHandle = (short)(((short)data[ 4] & 0xff) | 
                                  (((short)data[ 5] & 0x0f) << 8));
        remoteAddress = (long)(((long)data[ 6]) & 0xffL)       | 
                              (((long)data[ 7]) & 0xffL) <<  8 | 
                              (((long)data[ 8]) & 0xffL) << 16 | 
                              (((long)data[ 9]) & 0xffL) << 24 | 
                              (((long)data[10]) & 0xffL) << 32 | 
                              (((long)data[11]) & 0xffL) << 40;
        remoteAddressString = BluetoothAddress.LongToStandardRepr(remoteAddress);
        //linkType = data[12];
        //encryptionMode = data[13];
        identifierCount = 1;
        linkState = L2CAP_LINK_OPEN;
        linkMTU = hciDriver.get_ACL_MTU();
        hciDriver.registerL2CAPLink(this);
    }

    public void close() {
        if (linkState == L2CAP_LINK_OPEN) {
            linkState = L2CAP_LINK_CLOSED;
            try {
                hciDriver.send_HCI_LC_Disconnect(connectionHandle);
            }
            catch (HCIException e) {
                wasDisconnected();
            }
        }
    }

    /**
     * Called by HCITransport when a Disconnect Complete Event was received.
     * This method closes all L2CAPChannels associated with this link, 
     * and ensures a proper cleanup.
     */
    public void wasDisconnected() {
        linkState = L2CAP_LINK_CLOSED;
        hciDriver.unregisterL2CAPLink(this);
        for (int i = 0; i < channels.length; ++i) {
            if (channels[i] != null) {
                channels[i].channelState = L2CAPChannel.CLOSED;
                channels[i] = null;
            }
        }
    }

    /**
     * Receives, parses, and dispatches HCI Data Packets.
     * @param data An ACL Data Packet received by the HCITransport.
     */
    public void receiveData(byte[] data) {
        //Debug.println(1, "L2CAPLink: in receiveData");
        // Perform L2CAP packet reassembly
        switch ((byte)(data[2] & 0x30) >>> 4) {
            case L2CAP_PACKET_BOUNDRY_FLAG_FIRST:
                packetBufferChannelID = (short)(((short)data[8] & 0xff) << 8 | 
                                                ((short)data[7] & 0xff));
                int packetLength = (short)(((short)data[6] & 0xff) << 8 | 
                                           ((short)data[5] & 0xff));
                packetBuffer = new byte[packetLength];
                packetBufferIndex = data.length - 9;
                System.arraycopy(data, 9, packetBuffer, 0, packetBufferIndex);
                break;
            case L2CAP_PACKET_BOUNDRY_FLAG_CONTINUING:
                int length = data.length - 5;
                if (length > packetBuffer.length - packetBufferIndex) {
                    length = packetBuffer.length - packetBufferIndex;
                    System.err.println("L2CAPLink: Size of continuing " +
                                       "packet too big.");
                }
                System.arraycopy(data, 5, packetBuffer, packetBufferIndex, length);
                packetBufferIndex += length;
                break;
            default:
                System.err.println("L2CAPLink: Received Packet with invalid " +
                                   "Boundry Flag : " + Debug.printByteArray(data));
        }
        if (packetBufferIndex == packetBuffer.length) {
            if (packetBufferChannelID == 1) { // signalling channel
                Debug.println(9, "L2CAPLink: Received signalling packet");
                dispatchSignallingPacket(packetBuffer);
            } else if (packetBufferChannelID == 2) { // broadcast channel
                Debug.println(2, "L2CAPLink: Received broadcast packet");
                dispatchBroadcastL2CAPPacket(packetBuffer);
            } else {
                L2CAPChannel channel = channels[packetBufferChannelID - 
                                                NUM_OF_RESERVED_CHANNELS];
                if (channel != null) {
                    Debug.println(9, "L2CAPLink: Received Packet for channel " + 
                                  channel.localChannelID + " from " + 
                                  remoteAddressString + " channel " + 
                                  channel.remoteChannelID);
                    channel.receiveL2CAPPacket(packetBuffer);
                } else System.err.println("L2CAPLink: Unable to deliver packet. " + 
                                          "Local channel " + packetBufferChannelID +
                                          " not found. " + 
                                          Debug.printByteArray(packetBuffer));
            }
        }
        //Debug.println(1, "L2CAPLink: out of receiveData");
    }

    public void sendL2CAPPacket(L2CAPChannel channel, byte[] data) 
                                throws IOException {
        sendL2CAPPacket(channel.remoteAddress, channel.remoteChannelID, data);
    }

    /** 
     * This function queues the given data for transmission on this link, 
     * performing also segmentation when needed.
     * @param data An array of bytes containing the data to be sent.
     */
    public void sendL2CAPPacket(long remoteAddress, short remoteChannelID, 
                                byte[] data) throws IOException {
        if (linkState != L2CAP_LINK_OPEN) 
            throw new IOException("Failed to queue L2CAP data for transmission. " + 
                                  "L2CAPLink is not open.");
        
        Debug.println(2, "L2CAPLink: queuing data for transmission to " + 
                      remoteAddressString + " channel " + remoteChannelID);

        // make sure that data is smaller than 64kb
        if (data.length > 65535)
            throw new IOException("Data buffer is too big! An L2CAP packet cannot " +
                                  "contain more than 65535 bytes of data.");

        /*
         * BEWARE, SEGMENTATION CODE IS BUGGY!!!!!!!!!!!!!!
         */
        
        // length of data to transmit
        int bytes_to_send = data.length;
        // length of already sent data
        int bytes_sent = 0;
        // to get neat mtu we must subtract the size of the L2CAP header from the acl mtu
        int neat_mtu = (int)linkMTU - L2CAP_HDR_LEN;
        // number of the fragment we are actually sending 
        int frag_num = 1;

        while (bytes_to_send > 0) {
            // does this stupid language have a min function for shorts?
            short l2cap_payload_size = (short)(neat_mtu < bytes_to_send ? neat_mtu : bytes_to_send);
            short hci_payload_size   = (short)(l2cap_payload_size + L2CAP_HDR_LEN);
            byte[] hciPacket  = new byte[hci_payload_size + HCI_ACL_HDR_LEN];
            byte boundary_flag = (frag_num == 1 ? FIRST_PACKET : CONTINUING_PACKET);

            // hci packet header
            hciPacket[0] = 0x02; // ACL packet type
            hciPacket[1] = (byte)((connectionHandle) & 0xff);
            hciPacket[2] = (byte)(((connectionHandle >>> 8) & 0x0f) | 
                                  ((boundary_flag & 0x03) << 4) | 
                                  ((H2C_NO_BROADCAST & 0x03) << 6));
            hciPacket[3] = (byte)((hci_payload_size) & 0xff);
            hciPacket[4] = (byte)((hci_payload_size >> 8) & 0xff);
            
            // l2cap (connection-oriented) packet header
            hciPacket[5] = (byte)((l2cap_payload_size) & 0xff);
            hciPacket[6] = (byte)((l2cap_payload_size >> 8) & 0xff);
            hciPacket[7] = (byte)((remoteChannelID) & 0xff);
            hciPacket[8] = (byte)((remoteChannelID >> 8) & 0xff);

            // l2cap payload
            System.arraycopy(data, bytes_sent, hciPacket, 
                             HCI_ACL_HDR_LEN + L2CAP_HDR_LEN, 
                             l2cap_payload_size);
            
            try { 
                hciDriver.send_HCI_Data_Packet(hciPacket); 
            }
            catch (HCIException e) { 
                throw new IOException("L2CAPLink: Error sending L2CAP packet." + e); 
            }

            bytes_sent += l2cap_payload_size;
            bytes_to_send -= l2cap_payload_size;
            ++frag_num;
        }
        
        Debug.println(9, "L2CAPLink: data successfully queued for transmission in " +
                      (frag_num - 1) +" fragments.");
    }

    /**
     * Process packets received for the L2CAP Signalling Channel.
     * @param packet Byte Array containing a L2CAP Packet.
     */
    private void dispatchSignallingPacket(byte[] packet) {
        Debug.println(9, "L2CAPLink: Received Packet on the Signalling Channel from " +
                      remoteAddressString);
        switch (packet[0]) {
            case L2CAP_COMMAND_REJECT:
                receive_L2CAP_Command_Reject(packet);
                break;
            case L2CAP_CONNECTION_REQUEST:
                receive_L2CAP_Connection_Request(packet);
                break;
            case L2CAP_CONNECTION_RESPONSE:
                receive_L2CAP_Connection_Response(packet);
                break;
            case L2CAP_CONFIGURE_REQUEST:
                receive_L2CAP_Configuration_Request(packet);
                break;
            case L2CAP_CONFIGURE_RESPONSE:
                receive_L2CAP_Configuration_Response(packet);
                break;
            case L2CAP_DISCONNECTION_REQUEST:
                receive_L2CAP_Disconnection_Request(packet);
                break;
            case L2CAP_DISCONNECTION_RESPONSE:
                receive_L2CAP_Disconnection_Response(packet);
                break;
            case L2CAP_ECHO_REQUEST:
                receive_L2CAP_Echo_Request(packet);
                break;
            case L2CAP_ECHO_RESPONSE:
                receive_L2CAP_Echo_Response(packet);
                break;
            case L2CAP_INFORMATION_REQUEST:
                receive_L2CAP_Information_Request(packet);
                break;
            case L2CAP_INFORMATION_RESPONSE:
                receive_L2CAP_Information_Response(packet);
                break;
            default:
                receive_L2CAP_Unknown_Signalling_Packet(packet);
        }
    }

    private void receive_L2CAP_Command_Reject(byte[] packet) {
        System.err.println("L2CAPLink: Received Command Reject Signal from " + 
                           remoteAddressString + " : " + Debug.printByteArray(packet));
    }

    private void receive_L2CAP_Connection_Request(byte[] packet) {
        Debug.println(1, "L2CAPLink: Received Connection Request Signal from " + 
                      remoteAddressString);
        short psm       = (short)(((short)packet[5] & 0xff) << 8 | 
                                  ((short)packet[4] & 0xff));
        short remoteCID = (short)(((short)packet[7] & 0xff) << 8 | 
                                  ((short)packet[6] & 0xff));
        short localCID  = -1;
        for (short i = 0; i < channels.length; ++i) {
            if (channels[i] == null) {
                localCID = i;
                break;
            }
        }
        int resultCode = 0x0000;
        short localChannelID = (short)(localCID + NUM_OF_RESERVED_CHANNELS);
        Debug.println(1, "L2CAP: localCID is: " + localCID);
        if (localCID == -1) {
            resultCode = 0x0004; /* Error: No Resources */
        } else {
            Debug.println(1, "L2CAP: 2fix resolveAndCreateL2CAPChannel");
            /*
            L2CAPChannel channel = sddb.resolveAndCreateL2CAPChannel(psm, 
                                          this, localChannelID, remoteCID);
            if (channel == null) {
                resultCode = 0x0002; // Error: No PSM
            } else {
                channels[localCID] = channel;
                channel.channelState = L2CAPChannel.CLOSED;
                channel.l2capSender = this;
                channel.localChannelID = localChannelID;
                channel.remoteChannelID = remoteCID;
            }
            */
        }
        byte[] connectionResponse = {
            0x03, 
            packet[1], 
            0x08, 
            0x00, 
            (byte)(localChannelID & 0xff), 
            (byte)((localChannelID >> 8) & 0xff), 
            packet[6], 
            packet[7], 
            (byte)((resultCode) & 0xff), 
            (byte)((resultCode >> 8) & 0xff), 
            0x00, 
            0x00 
        };
        try {
            Debug.println(3, "L2CAP: Sending Connection Response Signal to " + 
                          remoteAddressString);
            sendL2CAPPacket(remoteAddress, (short)0x0001, connectionResponse);
        }
        catch (IOException e) {
            this.close();
        }
    }

    private void receive_L2CAP_Connection_Response(byte[] packet) {
        Debug.println(3, "L2CAP: Received Connection Response Signal from " + 
                      remoteAddressString);
        short remoteCID = (short)(((short)packet[ 5] & 0xff) << 8 | 
                                  ((short)packet[ 4] & 0xff));
        short localCID  = (short)(((short)packet[ 7] & 0xff) << 8 | 
                                  ((short)packet[ 6] & 0xff));
        short result    = (short)(((short)packet[ 9] & 0xff) << 8 | 
                                  ((short)packet[ 8] & 0xff));
        short status    = (short)(((short)packet[11] & 0xff) << 8 | 
                                  ((short)packet[10] & 0xff));
        L2CAPChannel channel = channels[localCID - NUM_OF_RESERVED_CHANNELS];
        if (channel != null) {
            channel.remoteChannelID = remoteCID;
            channel.channelState = L2CAPChannel.CONFIG;
            try { send_L2CAP_Configuration_Request(channel); }
            catch (IOException e) { this.close(); }
        }
    }

    private void receive_L2CAP_Configuration_Request(byte[] packet) {
        Debug.println(3, "L2CAP: Received Configuration Request Signal from " + 
                      remoteAddressString);
        short localCID = (short)(((short)packet[5] & 0xff) << 8 | 
                                 ((short)packet[4] & 0xff));
        short optionLength = (short)((((short)packet[3] & 0xff) << 8 | 
                                      ((short)packet[2] & 0xff)) - 4);
        /* TODO: parse configuration options in a proper way... 
         * this implementation just acknowledges but never challenges 
         * the proposed authentication */
        L2CAPChannel channel = channels[localCID - NUM_OF_RESERVED_CHANNELS];
        if (channel != null) {
            byte[] configurationResponse = new byte[/*optionLength+*/ 10];
            configurationResponse[0] = 0x05;
            configurationResponse[1] = packet[1];
            configurationResponse[2] = (byte)((/*optionLength+*/ 6) & 0xff);
            configurationResponse[3] = (byte)(((/*optionLength+*/ 6) >> 8) & 0xff);
            configurationResponse[4] = (byte)((channel.remoteChannelID) & 0xff);
            configurationResponse[5] = (byte)((channel.remoteChannelID >> 8) & 0xff);
            configurationResponse[6] = 0x00; //no flags
            configurationResponse[7] = 0x00; //no flags
            configurationResponse[8] = 0x00; //result ok
            configurationResponse[9] = 0x00; //result ok
            //System.arraycopy(packet,8,configurationResponse,10,optionLength);
            //configurationResponse[12]=0x00;//new mtu
            //configurationResponse[13]=(byte) 0xff;//new mtu                                                
            try {
                Debug.println(3, "L2CAP: Sending Cofiguration Response Signal to " + 
                              remoteAddressString);
                sendL2CAPPacket(remoteAddress, (short)0x0001, configurationResponse);
                if (channel.channelState == L2CAPChannel.CLOSED) {
                    channel.channelState = L2CAPChannel.CONFIG;
                    send_L2CAP_Configuration_Request(channel);
                }
            }
            catch (IOException e) { this.close(); }
        }
    }

    private void receive_L2CAP_Configuration_Response(byte[] packet) {
        Debug.println(3, "L2CAP: Received Configuration Response Signal from " +
                      remoteAddressString);
        short localCID = (short)(((short)packet[5] & 0xff) << 8 | 
                                 ((short)packet[4] & 0xff));
        short result = (short)(((short)packet[9] & 0xff) << 8 | 
                               ((short)packet[8] & 0xff));
        if (result == 0) {
            L2CAPChannel channel = channels[localCID - NUM_OF_RESERVED_CHANNELS];
            if (channel != null) {
                if (channel.channelState == L2CAPChannel.CLOSED) 
                    channel.channelState = L2CAPChannel.CONFIG;
                else if (channel.channelState == L2CAPChannel.CONFIG) 
                    channel.channelState = L2CAPChannel.OPEN;
            }
        }
        /* TODO: add possible challenge to the response.
         * this implementation just nods and agrees... */
    }

    private void receive_L2CAP_Disconnection_Request(byte[] packet) {
        Debug.println(3, "L2CAP: Received Disconnection Request Signal from " + 
                      remoteAddressString);
        short localCID = (short)(((short)packet[5] & 0xff) << 8 |
                                 ((short)packet[4] & 0xff));
        /*
        short remoteCID=(short)(((short) packet[7] & 0xff) << 8 | 
                                ((short) packet[6]& 0xff));
        */
        L2CAPChannel channel = channels[localCID - NUM_OF_RESERVED_CHANNELS];
        channels[localCID - NUM_OF_RESERVED_CHANNELS] = null;
        
        if (channel != null) {
            channel.wasDisconnected();
        }
        
        byte[] disconnectionResponse = { 
            0x07, packet[1], 0x04, 0x00, 
            packet[4], packet[5], packet[6], packet[7] 
        };
        
        try {
            Debug.println(3, "L2CAP: Sending Disconnection Response Signal to " + 
                          remoteAddressString);
            sendL2CAPPacket(remoteAddress, (short)0x0001, disconnectionResponse);
        }
        catch (IOException e) {
            wasDisconnected();
        }
    }

    private void receive_L2CAP_Disconnection_Response(byte[] packet) {
        Debug.println(3, "L2CAP: Received Disconnection Response Signal from " + 
                      remoteAddressString);
        short localCID = (short)(((short)packet[8] & 0xff) << 8 | 
                                 ((short)packet[7] & 0xff));
        L2CAPChannel channel = channels[localCID - NUM_OF_RESERVED_CHANNELS];
        channels[localCID - NUM_OF_RESERVED_CHANNELS] = null;
        if (channel != null) { channel.wasDisconnected(); }
    }

    private void receive_L2CAP_Echo_Request(byte[] packet) {
        Debug.println(3, "L2CAP: Received Echo Request Signal from " + 
                      remoteAddressString);
    }

    private void receive_L2CAP_Echo_Response(byte[] packet) {
        Debug.println(3, "L2CAP: Received Echo Response Signal from " + 
                      remoteAddressString);
    }

    private void receive_L2CAP_Information_Request(byte[] packet) {
        Debug.println(3, "L2CAP: Received Information Request Signal from " + 
                      remoteAddressString);
    }

    private void receive_L2CAP_Information_Response(byte[] packet) {
        Debug.println(3, "L2CAP: Received Information Response Signal from " + 
                      remoteAddressString);
    }

    private void receive_L2CAP_Unknown_Signalling_Packet(byte[] packet) {
        System.err.println("L2CAP: Received unknown signalling packet:" + 
                           Debug.printByteArray(packet));
        send_L2CAP_Command_Reject();
    }

    public void send_L2CAP_Command_Reject() {
        Debug.println(3, "L2CAP: Sending Command Rejected Signal from " + 
                      remoteAddressString);
    }

    public void send_L2CAP_Connection_Request(L2CAPChannel channel, short psm) 
                                              throws HCIException, IOException {
        channel.channelState = L2CAPChannel.CLOSED;
        
        for (short i = 0; i < channels.length; i++) {
            if (channels[i] == null) {
                channels[i] = channel;
                channel.channelState = L2CAPChannel.CLOSED;
                channel.remoteAddress = remoteAddress;
                channel.l2capSender = this;
                channel.localChannelID = (short)(i + NUM_OF_RESERVED_CHANNELS);
                break;
            }
        }
        if (channel.localChannelID == -1)
            throw new HCIException("L2CAP Connection Request failed. " +
                                   "No Local Channels available.");
        
        byte[] connectionRequest = {
            0x02, 
            identifierCount++, 
            0x04, //fixed length
            0x00, //fixed length
            (byte)((psm) & 0xff), 
            (byte)((psm >> 8) & 0xff), 
            (byte)((channel.localChannelID) & 0xff),
            (byte)((channel.localChannelID >> 8) & 0xff)
        };
        
        Debug.println(3, "L2CAP: Sending Connection Request Signal to " + 
                      remoteAddressString);
        sendL2CAPPacket(remoteAddress, (short)0x0001, connectionRequest);
    }

    public void send_L2CAP_Configuration_Request(L2CAPChannel channel) 
                                                 throws IOException {
        byte[] configurationRequest = {
            0x04, 
            identifierCount++, 
            0x04, //fixed length
            0x00, //fixed length
            (byte)((channel.remoteChannelID) & 0xff),
            (byte)((channel.remoteChannelID >> 8) & 0xff),
            0x00, //no flags
            0x00
        };
        Debug.println(3, "L2CAP: Sending Configuration Request Signal to " + 
                      remoteAddressString);
        sendL2CAPPacket(remoteAddress, (short)0x0001, configurationRequest);
    }

    public void send_L2CAP_Disconnection_Request(L2CAPChannel channel) 
                                                 throws IOException {
        byte[] disconnectionRequest = {
            0x06, identifierCount++, 0x04, 0x00, 
            (byte)((channel.remoteChannelID) & 0xff), 
            (byte)((channel.remoteChannelID >> 8) & 0xff), 
            (byte)((channel.localChannelID) & 0xff), 
            (byte)((channel.localChannelID >> 8) & 0xff)
        };
        channel.channelState = L2CAPChannel.CLOSED;
        Debug.println(3, "L2CAP: Sending Disconnection Request Signal to " + remoteAddressString);
        sendL2CAPPacket(remoteAddress, (short)0x0001, disconnectionRequest);
    }

    public void send_L2CAP_Echo_Request(L2CAPChannel channel) {
        /*
        Debug.println(1, "L2CAP: Sending Echo Request to " + remoteAddressString);
        */
        /* TODO: send echo request */
    }

    public void send_L2CAP_Information_Request(L2CAPChannel channel) {
        /* 
        Debug.println(1, "L2CAP: Sending Information Request to " + remoteAddressString);
        */
        /* TODO: send information request */
    }

    public void connectL2CAPChannel(L2CAPChannel channel, short psm) 
                                    throws HCIException {
        try {
            send_L2CAP_Connection_Request(channel, psm);
        }
        catch (IOException e) {
            throw new HCIException("L2CAPChannel Connection Request Failed.");
        }        
        int timeout = 0;
        while (channel.channelState != L2CAPChannel.OPEN) {
            try {
                Thread.sleep(1000);
                timeout++;
            }
            catch (InterruptedException e) { }
            if (timeout > 50)
                throw new HCIException("L2CAPChannel Connection Request timed out.");
        }
    }

    public static void connectL2CAPBroadcastChannel(ASBChannel channel) throws HCIException {
        synchronized (asbChannelsLock) {
            /* make sure that a group with PSM = channel.psm does not exist yet */
            Short psm = new Short(channel.psm);
            if (asbChannelsByPSM.containsKey(psm)) {
                throw new RuntimeException("ASB channel with PSM = " + psm + 
                                           " already exists.");
            }
        
            /* create group */
            channel.localChannelID = -1;
            for (short i = 0; i < asbChannels.length; ++i) {
                if (asbChannels[i] == null) {
                    channel.localChannelID = i;
                    break;
                }
            }
            if (channel.localChannelID == -1) {
                throw new RuntimeException("Cannot create more than " + 
                                           asbChannels.length + 
                                           " ASB channels.");
            }
            
            asbChannels[channel.localChannelID] = channel;
            asbChannelsByPSM.put(psm, channel);
        }
        
        channel.remoteChannelID = BROADCAST_CHANNEL;
    }
    
    public static void sendL2CAPBroadcastPacket(ASBChannel channel, 
                                                byte[] data) throws IOException { 
        short psm = channel.psm;
        int link_mtu = channel.hciDriver.get_ACL_MTU();
        
        /*
        Debug.println(2, "L2CAPLink: queuing data (packet " + bcast_packet_sent + ", " +
                      data.length + " bytes) for broadcast transmission on PSM " + psm);
        */

        // reserved space on each packet for statistics
        int reserved = 4;

        // make sure that data is smaller than 64kb
        if (data.length > 65535 - (L2CAP_HDR_LEN + PSM_FIELD_SIZE + reserved))
            throw new IOException("Data buffer is too big! An L2CAP packet cannot " +
                                  "contain more than 65535 bytes of data.");

        /* TODO: assert that channel is a valid/registered ASB channel */
        /*        
        if (!asbChannelsByPSM.containsKey(psm)) {
            throw new RuntimeException("ASB channel with PSM = " + psm + 
                                       " is not registered.");
        }
        if (asbChannelsByPSM.get(new Short(psm)) != channel) {
            throw new RuntimeException("The given ASB channel " +
                                       "is not the one registered " +
                                       "for PSM = " + psm);
        }
         */
       
        /*
         * SEGMENTATION CODE IS BUGGY!!!!!!!!!!!!!!
         */
        
        // length of data to transmit
        int bytes_to_send = data.length;
        // length of already sent data
        int bytes_sent = 0;
        // to get neat mtu we must subtract the size of the L2CAP header plus
        // the psm field from the acl mtu and 4 bytes for sequence number
        int neat_mtu = link_mtu - (L2CAP_HDR_LEN + PSM_FIELD_SIZE + reserved);
        // number of the fragment we are actually sending
        int frag_num = 1;

        /*
        Debug.println(1, "L2CAPBroadcast: neat_mtu is " + neat_mtu + "bytes");
        Debug.println(1, "L2CAPBroadcast: link_mtu is " + link_mtu + "bytes");
        Debug.println(1, "L2CAPBroadcast: sending " + bytes_to_send + "bytes");
        */
        
        while (bytes_to_send > 0) {
            // does this stupid language have a min function for shorts?
            short appl_payload_size  = (short)(neat_mtu < bytes_to_send ? neat_mtu : bytes_to_send);
            short l2cap_payload_size = (short)(appl_payload_size + PSM_FIELD_SIZE + reserved);
            short hci_payload_size   = (short)(l2cap_payload_size + L2CAP_HDR_LEN);
            byte[] hciPacket  = new byte[hci_payload_size + HCI_ACL_HDR_LEN];
            byte boundary_flag = (frag_num == 1 ? FIRST_PACKET : CONTINUING_PACKET);
            
            /*
            Debug.println(1, "L2CAPBroadcast: appl_payload_size is " + appl_payload_size + "bytes");
            Debug.println(1, "L2CAPBroadcast: l2cap_payload_size is " + l2cap_payload_size + "bytes");
            Debug.println(1, "L2CAPBroadcast: hci_payload_size is " + hci_payload_size + "bytes");
            Debug.println(1, "L2CAPBroadcast: hciPacket is " + hciPacket.length + "bytes");
            */

            // hci packet header
            hciPacket[ 0] = 0x02; // ACL packet type
            hciPacket[ 1] = (byte)((ASBConnectionHandle) & 0xff);
            hciPacket[ 2] = (byte)(((ASBConnectionHandle >>> 8) & 0x0f) | 
                                   ((boundary_flag & 0x03) << 4) | 
                                   ((H2C_ACTIVE_SLAVE_BROADCAST & 0x03) << 6));
            hciPacket[ 3] = (byte)((hci_payload_size) & 0xff);
            hciPacket[ 4] = (byte)((hci_payload_size >> 8) & 0xff);
            
            // l2cap (connection-oriented) packet header
            hciPacket[ 5] = (byte)((l2cap_payload_size) & 0xff);
            hciPacket[ 6] = (byte)(((l2cap_payload_size) >> 8) & 0xff);
            hciPacket[ 7] = (byte)((BROADCAST_CHANNEL) & 0xff);
            hciPacket[ 8] = (byte)((BROADCAST_CHANNEL >> 8) & 0xff);
            hciPacket[ 9] = (byte)((psm) & 0xff);
            hciPacket[10] = (byte)((psm >> 8) & 0xff);            
            hciPacket[11] = (byte)((bcast_packet_sent) & 0xff);
            hciPacket[12] = (byte)((bcast_packet_sent >>  8) & 0xff);
            hciPacket[13] = (byte)((bcast_packet_sent >> 16) & 0xff);
            hciPacket[14] = (byte)((bcast_packet_sent >> 24) & 0xff);

            // l2cap payload
            System.arraycopy(data, bytes_sent, hciPacket, 
                             HCI_ACL_HDR_LEN + L2CAP_HDR_LEN + PSM_FIELD_SIZE + reserved, 
                             appl_payload_size);
            
            try { 
                channel.hciDriver.send_HCI_Data_Packet(hciPacket); 
            }
            catch (HCIException e) { 
                throw new IOException("L2CAPLink: Error sending broadcast L2CAP packet." + e); 
            }

            bytes_sent += appl_payload_size;
            bytes_to_send -= appl_payload_size;
            ++frag_num;
            bcast_packet_sent++;
        }
        
        /*
        Debug.println(9, "L2CAPLink: data successfully queued for broadcast transmission in " +
                      (frag_num - 1) +" fragments.");
        */
    }
    
    public static void sendL2CAPBroadcastPacketGOODVERSION(ASBChannel channel, 
                                                byte[] data) throws IOException { 
        short psm = channel.psm;
        int link_mtu = channel.hciDriver.get_ACL_MTU();
        
        Debug.println(2, "L2CAPLink: queuing data for broadcast transmission on PSM " + psm);

        // make sure that data is smaller than 64kb
        if (data.length > 65535)
            throw new IOException("Data buffer is too big! An L2CAP packet cannot " +
                                  "contain more than 65535 bytes of data.");

        /* TODO: assert that channel is a valid/registered ASB channel */
        /*        
        if (!asbChannelsByPSM.containsKey(psm)) {
            throw new RuntimeException("ASB channel with PSM = " + psm + 
                                       " is not registered.");
        }
        if (asbChannelsByPSM.get(new Short(psm)) != channel) {
            throw new RuntimeException("The given ASB channel " +
                                       "is not the one registered " +
                                       "for PSM = " + psm);
        }
         */
        
        /*
         * SEGMENTATION CODE IS BUGGY!!!!!!!!!!!!!!
         */

        // length of data to transmit
        int bytes_to_send = data.length;
        // length of already sent data
        int bytes_sent = 0;
        // to get neat mtu we must subtract the size of the L2CAP header plus
        // the psm field from the acl mtu
        int neat_mtu = link_mtu - (L2CAP_HDR_LEN + PSM_FIELD_SIZE);
        // number of the fragment we are actually sending 
        int frag_num = 1;

        while (bytes_to_send > 0) {
            // does this stupid language have a min function for shorts?
            short l2cap_payload_size = (short)(neat_mtu < bytes_to_send ? neat_mtu : bytes_to_send);
            short hci_payload_size   = (short)(l2cap_payload_size + L2CAP_HDR_LEN + PSM_FIELD_SIZE);
            byte[] hciPacket  = new byte[hci_payload_size + HCI_ACL_HDR_LEN];
            byte boundary_flag = (frag_num == 1 ? FIRST_PACKET : CONTINUING_PACKET);

            // hci packet header
            hciPacket[ 0] = 0x02; // ACL packet type
            hciPacket[ 1] = (byte)((ASBConnectionHandle) & 0xff);
            hciPacket[ 2] = (byte)(((ASBConnectionHandle >>> 8) & 0x0f) | 
                                   ((boundary_flag & 0x03) << 4) | 
                                   ((H2C_ACTIVE_SLAVE_BROADCAST & 0x03) << 6));
            hciPacket[ 3] = (byte)((hci_payload_size) & 0xff);
            hciPacket[ 4] = (byte)((hci_payload_size >> 8) & 0xff);
            
            // l2cap (connection-oriented) packet header
            hciPacket[ 5] = (byte)((l2cap_payload_size + PSM_FIELD_SIZE) & 0xff);
            hciPacket[ 6] = (byte)(((l2cap_payload_size + PSM_FIELD_SIZE) >> 8) & 0xff);
            hciPacket[ 7] = (byte)((BROADCAST_CHANNEL) & 0xff);
            hciPacket[ 8] = (byte)((BROADCAST_CHANNEL >> 8) & 0xff);
            hciPacket[ 9] = (byte)((psm) & 0xff);
            hciPacket[10] = (byte)((psm >> 8) & 0xff);

            // l2cap payload
            System.arraycopy(data, bytes_sent, hciPacket, 
                             HCI_ACL_HDR_LEN + L2CAP_HDR_LEN + PSM_FIELD_SIZE, 
                             l2cap_payload_size);
            
            try { 
                channel.hciDriver.send_HCI_Data_Packet(hciPacket); 
            }
            catch (HCIException e) { 
                throw new IOException("L2CAPLink: Error sending broadcast L2CAP packet." + e); 
            }

            bytes_sent += l2cap_payload_size;
            bytes_to_send -= l2cap_payload_size;
            ++frag_num;
        }
        
        Debug.println(9, "L2CAPLink: data successfully queued for broadcast transmission in " +
                      (frag_num - 1) +" fragments.");
    }
    
    private void dispatchBroadcastL2CAPPacket(byte[] packet) {
        Debug.println(9, "L2CAPLink: Received Packet on the Signalling Channel from " +
                      remoteAddressString);
        
        /* get psm */
        /* TODO: check if these offsets are ok */
        short psm = (short)((((short)packet[4]) & 0xff) |
                            (((short)packet[5]) & 0xff) << 8);
        
        /* find asb channel by psm */
        ASBChannel channel = (ASBChannel)asbChannelsByPSM.get(new Short(psm));
        
        /* retrieve data portion */
        int size = packet.length - 11;
        byte[] data = new byte[size];
        System.arraycopy(packet, 11, data, 0, size);
        
        /* deliver data */
        channel.receiveL2CAPPacket(data);
    }

    /** @see org.javabluetooth.stack.l2cap.L2CAPSender#closeL2CAPChannel(...) */
    public void closeL2CAPChannel(L2CAPChannel channel) {
        try { send_L2CAP_Disconnection_Request(channel); }
        catch (IOException e) { }
    }
}
/*
 * vim: et ts=4 sw=4
 */

