#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdint.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/sco.h>

#include "libBluezSCOConnection.h"

#define PACKET_BUFFER_SIZE 4096

#ifdef DEBUG
#define DPRINTF(args)	{ printf args ;  fflush(stdout); }
#define DPUTS(str)	{ puts(str); fflush(stdout); }
#else
#define DPRINTF(args)	do {} while(0);
#define DPUTS(str)	do {} while(0);
#endif

static int strjava2c(JNIEnv *env, jobject jstr, char **cstr);
static int is_valid_bluetooth_address(const char *address);


JNIEXPORT jint JNICALL Java_org_javabluetooth_stack_sco_SCOTransport_startSCO
  (JNIEnv *env, jobject obj, jstring address, jobject tx_queue)
{
	int fd;
	uint8_t buf[PACKET_BUFFER_SIZE];
	struct sockaddr_sco ss;
	char *remote = NULL;

	DPUTS("in startSCO()");

#ifdef DEBUG
	{
		jclass tmp_cls;
		tmp_cls = (*env)->FindClass(env, "java/lang/String");
		if (tmp_cls == NULL) {
			puts("tmp_cls (String) == NULL"); fflush(stdout);
		} else {
			int is;
			
			puts("got valid tmp_cls (String)"); fflush(stdout);
			
			is = (*env)->IsInstanceOf(env, address, tmp_cls);
			printf("tx_queue %s an instance of String\n", 
			       (is == JNI_TRUE ? "is" : "is not"));
			fflush(stdout);
		}
		
		tmp_cls = (*env)->FindClass(env, "org/javabluetooth/util/PacketQueue");
		if (tmp_cls == NULL) {
			puts("tmp_cls (PacketQueue) == NULL"); fflush(stdout);
		} else {
			int is;
			
			puts("got valid tmp_cls (PacketQueue)"); fflush(stdout);
			
			is = (*env)->IsInstanceOf(env, tx_queue, tmp_cls);
			printf("tx_queue %s an instance of PacketQueue\n", 
			       (is == JNI_TRUE ? "is" : "is not"));
			fflush(stdout);
		}
#endif
	
	/* Create SCO socket */
	fd = socket(AF_BLUETOOTH, SOCK_SEQPACKET, BTPROTO_SCO);
	if (fd < 0)
		return -1;

	DPRINTF(("SCO socket created: %d\n", fd));
	
	/* we use device ID 0 (hci0) for the moment */
	memset(&ss, 0, sizeof(ss));
	ss.sco_family = AF_BLUETOOTH;
	if (bind(fd, (struct sockaddr *)&ss, sizeof(ss)) < 0) {
		DPRINTF(("error with bind(2): %s (%d)\n", strerror(errno), errno));
		return -1;
	}

	strjava2c(env, address, &remote);
	if (remote == NULL || strlen(remote) != 17 || !is_valid_bluetooth_address(remote)) {
		/* TODO: raise an IllegalArgumentException */
	}
	
	memset(&ss, 0, sizeof(ss));
	ss.sco_family = AF_BLUETOOTH;
	baswap(&(ss.sco_bdaddr), strtoba(remote));
	if (connect(fd, (struct sockaddr *)&ss, sizeof(ss)) < 0) {
		DPRINTF(("error with connect(2): %s (%d)\n", strerror(errno), errno));
		return -1;
	}	
	
	/* The select loop: */
	for (;;) {
		int rr, max_fd;
		fd_set read_fdset, write_fdset;
		struct timeval timeout;
		int have_packet_to_send = 0;

		DPUTS("start of the loop");

		/* find out if we have a packet to send */
		{
			jboolean tx_queue_is_empty;
			jclass tx_queue_cls;
			jmethodID isEmpty;
	
			/* entering critical section */
			//puts("1"); fflush(stdout);
			//(*env)->MonitorEnter(env, tx_queue);
	
			//puts("2"); fflush(stdout);
			tx_queue_cls = (*env)->GetObjectClass(env, tx_queue);
			assert(tx_queue_cls != NULL);
			
			//puts("2a"); fflush(stdout);
			isEmpty = (*env)->GetMethodID(env, tx_queue_cls, "isEmpty", "()Z");
			assert(isEmpty != NULL);
			
			/* find out if the tx_queue list is empty */
			//puts("2b"); fflush(stdout);
			tx_queue_is_empty = (*env)->CallBooleanMethod(env, tx_queue, isEmpty);
			
			//puts("3"); fflush(stdout);
			have_packet_to_send = (tx_queue_is_empty == JNI_TRUE ? 0: 1);
			//printf("have_packet_to_send is: %d\n", have_packet_to_send); fflush(stdout);
			
			/* exiting critical section */
			//puts("4"); fflush(stdout);
			//(*env)->MonitorExit(env, tx_queue);
			
			//puts("5"); fflush(stdout);
		}
		
		//puts("100"); fflush(stdout);
		/* setup fd_set stuff for select */
		FD_ZERO(&read_fdset);
		FD_ZERO(&write_fdset);

		//puts("101"); fflush(stdout);
		max_fd = fd;
		FD_SET(fd, &read_fdset);
		if (have_packet_to_send)
			FD_SET(fd, &write_fdset);

		//puts("102"); fflush(stdout);
		/* setup 1 second timeout */
		timeout.tv_sec  = 1;
		timeout.tv_usec = 0;

		//puts("103"); fflush(stdout);
		/* blocking select with timeout */
		rr = select(max_fd + 1, &read_fdset, &write_fdset, NULL, &timeout);

		//puts("104"); fflush(stdout);
		/* handle select errors.
		 * if errno == EINTR we just retry select */
		if (rr < 0) {
			if (errno == EINTR) continue;
			return -2;
		}
		
		//puts("105"); fflush(stdout);		
		//printf("FD_ISSET(fd, &read_fdset) is: %d\n", FD_ISSET(fd, &read_fdset)); fflush(stdout);
		if (FD_ISSET(fd, &read_fdset)) {
			/* fd is ready for reading */
			int i;
			jbyteArray packet;
			jbyte *packet_buf;
			jclass obj_cls;
			jmethodID receiveData;

			/* read data from the socket */
			//puts("11"); fflush(stdout);
			rr = recv(fd, buf, sizeof(buf), 0);
			
			//puts("12"); fflush(stdout);
			if (rr < 0) return -3;

			/* create packet */
			//puts("13"); fflush(stdout);
			packet = (*env)->NewByteArray(env, rr);
			
			//puts("14"); fflush(stdout);
			packet_buf = (*env)->GetByteArrayElements(env, packet, 0);
			
			//puts("15"); fflush(stdout);
			for (i = 0; i < rr; ++i) { // optimize here
				packet_buf[i] = buf[i];
			}
			
			//puts("16"); fflush(stdout);
			(*env)->ReleaseByteArrayElements(env, packet, packet_buf, 0);
			
			//puts("17"); fflush(stdout);
			obj_cls = (*env)->GetObjectClass(env, obj);
			assert(obj_cls != NULL);
			
			//puts("17a"); fflush(stdout);
			receiveData = (*env)->GetMethodID(env, obj_cls, "receiveData", "([B)V");
			assert(receiveData != NULL);
			
			/* send packet to the HCIPacketListener */
			//puts("17b"); fflush(stdout);
			(*env)->CallVoidMethod(env, obj, receiveData, packet);

			//puts("18"); fflush(stdout);
		}

		/*
		printf("have_packet_to_send is: %d\n", have_packet_to_send);
		printf("FD_ISSET(fd, &write_fdset) is: %d\n", FD_ISSET(fd, &write_fdset));
		fflush(stdout);
		*/
		if (have_packet_to_send && FD_ISSET(fd, &write_fdset)) {
			/* fd is ready for writing */
			int i;
			jclass tx_queue_cls;
			jmethodID extract;
			jbyteArray packet;
			jbyte *packet_buf;
			jsize len;

			/* entering critical section */
			//puts("21"); fflush(stdout);
			//(*env)->MonitorEnter(env, tx_queue);
			
			//puts("21"); fflush(stdout);
			tx_queue_cls = (*env)->FindClass(env, "org/javabluetooth/util/PacketQueue");
			//tx_queue_cls = (*env)->GetObjectClass(env, tx_queue);
			assert(tx_queue_cls != NULL);
		
			//puts("21a"); fflush(stdout);
			extract = (*env)->GetMethodID(env, tx_queue_cls, "extract", "()[B");
			assert(extract != NULL);
	
			/* extract packet from packet queue */
			//puts("22"); fflush(stdout);
			packet = (*env)->CallObjectMethod(env, tx_queue, extract);
			/*
			if (packet == NULL) 
				puts("packet == NULL");
			else
				puts("got valid packet");
			fflush(stdout);
			*/

#if 0
			puts("23"); fflush(stdout);
			{
				jclass array_cls;
				jclass byte_cls;
				jclass byte_array_cls;
			
				array_cls = (*env)->FindClass(env, "java/lang/reflect/Array");
				if (array_cls == NULL) {
					puts("array_cls == NULL"); fflush(stdout);
				} else {
					int is;
					
					puts("got valid array_cls"); fflush(stdout);
					
					is = (*env)->IsInstanceOf(env, dummy, array_cls);
					printf("dummy %s an instance of Array\n", 
					        (is == JNI_TRUE ? "is" : "is not"));
					fflush(stdout);
				}
				
				byte_cls = (*env)->FindClass(env, "java/lang/Byte");
				if (byte_cls == NULL) {
					puts("byte_cls == NULL\n"); fflush(stdout);
				} else {
					int is;
					
					puts("got valid byte_cls"); fflush(stdout);
					
					is = (*env)->IsInstanceOf(env, dummy, byte_cls);
					printf("dummy %s an instance of Byte\n", 
					       (is == JNI_TRUE ? "is" : "is not"));
					fflush(stdout);
				}
				

				byte_array_cls = (*env)->FindClass(env, "[java/lang/Byte");
				if (byte_array_cls == NULL) {
					puts("byte_array_cls == NULL"); fflush(stdout);
				} else {
					int is;
					
					puts("got valid byte_array_cls"); fflush(stdout);
					
					is = (*env)->IsInstanceOf(env, dummy, byte_array_cls);
					printf("dummy %s an array of Byte\n", 
					       (is == JNI_TRUE ? "is" : "is not"));
					fflush(stdout);
				}
				
			}
#endif

			/* copy packet on buf */
			//puts("24"); fflush(stdout);
			len = (*env)->GetArrayLength(env, packet);
			
			//puts("25"); fflush(stdout);
			assert(len < PACKET_BUFFER_SIZE);
			packet_buf = (*env)->GetByteArrayElements(env, packet, 0);
			
			//puts("26"); fflush(stdout);
			for (i = 0; i < len; ++i) { // optimize here
				buf[i] = packet_buf[i];
			}
			
			//puts("27"); fflush(stdout);
			(*env)->ReleaseByteArrayElements(env, packet, packet_buf, JNI_ABORT);
			
			/* write data to the socket */
			//puts("28"); fflush(stdout);
			//printf("len = %d\n", len);
			rr = send(fd, buf, len, 0);
			if (rr < 0) {
				printf("error with send(2): %s (%d)\n", strerror(errno), errno);
				fflush(stdout);
				return -4;
			}
			
			//puts("30"); fflush(stdout);
		}
	}

	close(fd);

	puts("exit from startBluez()"); fflush(stdout);
	
	return 0;
}

/*
 * translates a Java String to a dynamically allocated C string 
 */
static int strjava2c(JNIEnv *env, jobject jstr, char **cstr)
{
	jclass jstr_cls;
	jmethodID getBytes;
	jobject array;
	jbyte *buf;
	jsize len;

	assert(env != NULL);
	assert(jstr != NULL);
	assert(cstr != NULL);
	
	/* entering critical section */
	//(*env)->MonitorEnter(env, jstr);

	DPUTS("strjava2c: 1");
	jstr_cls = (*env)->GetObjectClass(env, jstr);
	assert(jstr_cls != NULL);
	
	DPUTS("strjava2c: 2");
	getBytes = (*env)->GetMethodID(env, jstr_cls, "getBytes", "()[B");
	assert(getBytes != NULL);
	
	DPUTS("strjava2c: 3");
	array = (*env)->CallObjectMethod(env, jstr, getBytes);
	assert(array != NULL);
	
#ifdef DEBUG
	{
		/* paranoid type checking ;-) */
		jclass array_cls;
		jclass byte_array_cls;
	
		array_cls = (*env)->FindClass(env, "java/lang/reflect/Array");
		if (array_cls == NULL) {
			DPUTS("array_cls == NULL");
		} else {
			int is;
			
			DPUTS("got valid array_cls");
			
			is = (*env)->IsInstanceOf(env, array, array_cls);
			DPRINTF(("array %s an instance of Array\n", 
				(is == JNI_TRUE ? "is" : "is not")));
		}
		
		byte_array_cls = (*env)->FindClass(env, "[java/lang/Byte");
		if (byte_array_cls == NULL) {
			DPUTS("byte_array_cls == NULL");
		} else {
			int is;
			
			DPUTS("got valid byte_array_cls");
			
			is = (*env)->IsInstanceOf(env, array, byte_array_cls);
			DPRINTF(("array %s an array of Byte\n", 
				(is == JNI_TRUE ? "is" : "is not")));
		}
	}
#endif
	
	DPUTS("strjava2c: 4");
	
	/* copy packet on buf */
	len = (*env)->GetArrayLength(env, array);
	
	/* this is a very important check!!! */
	if (sizeof(char) != sizeof(jbyte)) {
		fprintf(stderr, "Internal error with character size!!!\n");
		return -1;
	}

	/* allocate string for remote BD_ADDR */
	*cstr = (char *)malloc(len + 1);
	if (*cstr == NULL) {
		fprintf(stderr, "Internal error with malloc!!!\n");
		return -2;
	}
	
	DPUTS("strjava2c: str 4");
	
	/* access the internal representation of array */
	buf = (*env)->GetByteArrayElements(env, array, 0);
	assert(buf != NULL);
	
	/* copy the BD_ADDR address to the C string */
	strncpy(*cstr, buf, len - 1);
	
	/* null-terminate the C string */
	*cstr[len - 1] = '\0';
	
	DPUTS("strjava2c: str 5");
	(*env)->ReleaseByteArrayElements(env, array, buf, JNI_ABORT);
	
	DPRINTF(("C string is: %s\n", *cstr));
	
	/* exiting critical section */
	//(*env)->MonitorExit(env, jstr);
	
	DPUTS("strjava2c: out");

	/* success */
	return 0;
}

/*
 * returns:
 * 1 if the argument is a valid BD_ADDR-format Bluetooth address, e.g. 00:11:22:33:44:55
 * 0 if the argument is not a valid BD_ADDR-format Bluetooth address
 */
static int is_valid_bluetooth_address(const char *address)
{
	int i;

	if (address == NULL || strlen(address) != 17) {
		/* should we report this error (perhaps by printing something 
		 * on stderr and exiting or by simply raising an 
		 * IllegalArgumentException)? */
		return 0;
	}

	DPRINTF(("is_valid_bluetooth_address: checking %s\n", address));
	
	for (i = 0; i < 18; i += 3) {
		if (!isdigit(i) || !isdigit(i+1))
			return 0;
	}
	
	for (i = 2; i < 15; i += 3) {
		if (address[i] != ':')
			return 0;
	}
	
	return 1;
}

