/*
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED.
 * 
 * This file is part of the JavaBluetooth Stack.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

package org.javabluetooth.util;

import java.lang.String;
import java.lang.Integer;
import java.lang.Long;
//import org.javabluetooth.util.Debug;

/** 
 * This class contains methods for handling Bluetooth addresses. 
 * @author Mauro Tortonesi
 */
public class BluetoothAddress {
    long address;

    public BluetoothAddress(String addr) {
        if (isValidJSR82Repr(addr)) {
            this.address = JSR82ToLongRepr(addr);
        } else if (isValidStandardRepr(addr)) {
            this.address = StandardToLongRepr(addr);
        } else throw new IllegalArgumentException("Invalid address: " + addr + "!");
    }

    public long getLongRepr() {
        return address;
    }
    
    public String getStandardRepr() {
        return LongToStandardRepr(address);
    }
    
    public String getJSR82Repr() {
        return LongToJSR82Repr(address);
    }
   
    /* are we sure that this returns always the same hashCode? */
    public int hashCode() {
        return (new Long(address)).hashCode();
    }
    
    public boolean equals(Object obj) {
        try {
            BluetoothAddress ba = (BluetoothAddress)obj;
            return (this.address == ba.address);
        }
        catch (Exception e) {
            return false;
        }
    }
    
    /* checks if this is a valid JSR82 representation of a Bluetooth address */
    public static boolean isValidJSR82Repr(String JSR82Addr) {
        //Debug.println(11, "isValidJSR82Repr: testing \"" + JSR82Addr + "\"");
        if (JSR82Addr.length() != 12) {
            //Debug.println(11, "isValidJSR82Repr: returning false.");
            return false;
        }
        try {
            long tmp = Long.parseLong(JSR82Addr, 16);
            if (tmp < 0 || tmp > 0xFFFFFFFFFFFFL) {
                //Debug.println(11, "isValidJSR82Repr: returning false.");
                return false;
            }
        }
        catch (NumberFormatException e) {
            //Debug.println(11, "isValidJSR82Repr: returning false.");
            return false;
        }
        //Debug.println(11, "isValidJSR82Repr: returning true.");
        return true;
    }
    
    private static boolean isHexByte(String str) {
        if (str.length() != 2) {
            return false;
        }
        try {
            int tmp = Integer.parseInt(str, 16);
            if (tmp < 0 || tmp > 0xFF) {
                return false;
            }
        }
        catch (NumberFormatException e) {
            return false;
        }
        return true;
    }
    
    public static boolean isValidStandardRepr(String Addr) {
        //Debug.println(9, "isValidStandardRepr: testing \"" + Addr + "\"");
        String _addr = Addr.trim();
        if (_addr.length() != 17) {
            //Debug.println(9, "FALSE 1");
            return false;
        }
        for (int i = 0; i < 6; i += 3) {
            if (!isHexByte(_addr.substring(i, i + 2))) {
                //Debug.println(9, "FALSE 2 on " + _addr.substring(i, i + 2));
                return false;
            }
        }
        for (int i = 2; i < 5; i += 3) {
            if (_addr.charAt(i) != ':') {
                //Debug.println(9, "FALSE 3 on position " + i);
                return false;
            }
        }
        return true;
    }
    
    public static boolean isValidLongRepr(long addr) {
        if (addr < 0 || addr > 0xFFFFFFFFFFFFL) {
            return false;
        }
        return true;
    }
    
    public static String JSR82ToStandardRepr(String JSR82Addr) {
        if (!isValidJSR82Repr(JSR82Addr)) {
            throw new IllegalArgumentException("Invalid JSR82 Bluetooth address: " + JSR82Addr + "!");
        }
        String _addr = JSR82Addr.trim();
        //Debug.println(11, "JSR82ToStandardRepr: JSR82 address trimmed.");
        return _addr.substring( 0,  2) + ":" +
               _addr.substring( 2,  4) + ":";
        /*+
               _addr.substring( 4,  6) + ":" +
               _addr.substring( 6,  8) + ":" +
               _addr.substring( 8, 10) + ":" +
               _addr.substring(10, 12);
               */
    }
    
    public static String StandardToJSR82Repr(String Addr) {
        if (!isValidStandardRepr(Addr)) {
            throw new IllegalArgumentException("Invalid Bluetooth address: " + Addr + "!");
        }
        String _addr = Addr.trim();
        return _addr.substring( 0,  2) +
               _addr.substring( 3,  5) +
               _addr.substring( 6,  8) +
               _addr.substring( 9, 11) +
               _addr.substring(12, 14) +
               _addr.substring(15, 17);
    }
    
    public static long JSR82ToLongRepr(String JSR82Addr) {
        if (!isValidJSR82Repr(JSR82Addr)) {
            throw new IllegalArgumentException("Invalid JSR82 Bluetooth address: " + JSR82Addr + "!");
        }
        String _addr = JSR82Addr.trim();
        return (Long.parseLong(_addr.substring( 0,  2), 16) << 40) +
               (Long.parseLong(_addr.substring( 2,  4), 16) << 32) +
               (Long.parseLong(_addr.substring( 4,  6), 16) << 24) +
               (Long.parseLong(_addr.substring( 6,  8), 16) << 16) +
               (Long.parseLong(_addr.substring( 8, 10), 16) <<  8) +
               (Long.parseLong(_addr.substring(10, 12), 16));
    }
    
    public static long StandardToLongRepr(String Addr) {
        if (!isValidStandardRepr(Addr)) {
            throw new IllegalArgumentException("Invalid standard Bluetooth address: " + Addr + "!");
        }
        String _addr = Addr.trim();
        return (Long.parseLong(_addr.substring( 0,  2), 16) << 40) +
               (Long.parseLong(_addr.substring( 3,  5), 16) << 32) +
               (Long.parseLong(_addr.substring( 6,  8), 16) << 24) +
               (Long.parseLong(_addr.substring( 9, 11), 16) << 16) +
               (Long.parseLong(_addr.substring(12, 14), 16) <<  8) +
               (Long.parseLong(_addr.substring(15, 17), 16));
    }
    
    public static String LongToStandardRepr(long Addr) {
        if (Addr < 0 || Addr > 0xFFFFFFFFFFFFL) {
            throw new IllegalArgumentException("Invalid Long Bluetooth address: " + Addr + "!");
        }
        return LongToUInt8HexString((Addr & 0xFF0000000000L) >> 40) + ":" +
               LongToUInt8HexString((Addr & 0x00FF00000000L) >> 32) + ":" +
               LongToUInt8HexString((Addr & 0x0000FF000000L) >> 24) + ":" +
               LongToUInt8HexString((Addr & 0x000000FF0000L) >> 16) + ":" +
               LongToUInt8HexString((Addr & 0x00000000FF00L) >>  8) + ":" +
               LongToUInt8HexString((Addr & 0x0000000000FFL));
    }
    public static String LongToJSR82Repr(long Addr) {
        String JSR82Addr = Long.toHexString(Addr).toUpperCase();
        while (JSR82Addr.length() < 12) {
            JSR82Addr = "0" + JSR82Addr;
        }
        return JSR82Addr;
    }
    
    /* these fucking java strings suck badly. we have to reimplement sprintf behaviour */
    private static String LongToUInt8HexString(long l) {
        if (l < 0 || l > 0xFF) {
            throw new IllegalArgumentException("Internal error with LongToUInt8HexString");
        }
        String conv = new String("0123456789ABCDEF");
        char[] result = new char[2]; 
        result[0] = conv.charAt((int)((l & 0xF0) >> 4));
        result[1] = conv.charAt((int)((l & 0x0F)));
        return new String(result);
    }
}
/*
 * vim: et ts=4 sw=4
 */

