/*
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED.
 * 
 * This file is part of the JavaBluetooth Stack.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */
package org.javabluetooth.util;

import java.util.LinkedList;
import org.javabluetooth.util.Debug;

/** 
 * This class implements a simple FIFO packet queue.
 * @author Mauro Tortonesi
 */
public class PacketQueue {
    LinkedList the_queue;
    int packets_in_queue;
    
    public static void main(String[] args) {
        PacketQueue pq = new PacketQueue();
        
        byte[] test1 = { (byte)0x01, (byte)0x23, (byte)0x45, (byte)0x67, (byte)0x89 };
        byte[] test2 = { (byte)0xab, (byte)0xcd, (byte)0xef };
        byte[] test3 = { (byte)0xde, (byte)0xad, (byte)0xbe, (byte)0xef };

        System.out.println("starting test 1");
        System.out.println("size is (0 expected): " + pq.size());
        System.out.println("isEmpty is (true expected): " + pq.isEmpty());
        
        System.out.println("starting test 2");
        pq.insert(test1);
        System.out.println("size is (1 expected): " + pq.size());
        System.out.println("isEmpty is (false expected): " + pq.isEmpty());
        
        System.out.println("starting test 3");
        pq.insert(test2);
        System.out.println("size is (2 expected): " + pq.size());
        System.out.println("isEmpty is (false expected): " + pq.isEmpty());
        
        System.out.println("starting test 4");
        byte[] dummy = pq.extract();
        System.out.println("size is (1 expected): " + pq.size());
        System.out.println("isEmpty is (false expected): " + pq.isEmpty());
        System.out.println("packet is (\"01 23 45 67 89\" expected): " + 
                           Debug.printByteArray(dummy));
        
        System.out.println("starting test 5");
        pq.insert(test3);
        System.out.println("size is (2 expected): " + pq.size());
        System.out.println("isEmpty is (false expected): " + pq.isEmpty());
        
        System.out.println("starting test 6");
        dummy = pq.extract();
        System.out.println("size is (1 expected): " + pq.size());
        System.out.println("isEmpty is (false expected): " + pq.isEmpty());
        System.out.println("packet is (\"ab cd ef\" expected): " + 
                           Debug.printByteArray(dummy));
        
        System.out.println("starting test 7");
        dummy = pq.extract();
        System.out.println("size is (0 expected): " + pq.size());
        System.out.println("isEmpty is (true expected): " + pq.isEmpty());        
        System.out.println("packet is (\"de ad be ef\" expected): " + 
                           Debug.printByteArray(dummy));
    }
    
    public PacketQueue() {
        the_queue = new LinkedList();
        packets_in_queue = 0;
    }

    /**
     * Extracts a packet from the queue.
     * @return the first packet inserted in the queue, or null if the queue is empty.
     */
    public synchronized byte[] extract() {
        if (packets_in_queue > 0) {
            int size = the_queue.size();
            
            if (size != packets_in_queue || size < 0) {
                System.err.println("Internal error in PackeQueue::extract()");
                System.exit(1);
            }
        
            byte[] packet = (byte[])the_queue.removeFirst();
            --packets_in_queue;
            
            return packet;
        }
        
        /* return null if the queue is empty */
        return null;
    }

    /**
     * Appends a packet to the queue.
     */
    public synchronized void insert(byte[] packet) {
        int size = the_queue.size();
        
        if (size != packets_in_queue || size < 0) {
            System.err.println("Internal error in PackeQueue::insert()");
            System.exit(1);
        }
        
        the_queue.addLast(packet);
        ++packets_in_queue;
    }

    /**
     * Checks if the queue is empty.
     * @return true if the queue is empty, and false otherwise.
     */
    public synchronized boolean isEmpty() {
        int size = the_queue.size();
        
        if (size != packets_in_queue || size < 0) {
            System.err.println("Internal error in PackeQueue::size()");
            System.exit(1);
        }

        return (size == 0);
    }
    
    /**
     * Returns the number of packets in the queue.
     * @return the number of packets in the queue.
     */
    public synchronized int size() {
        int size = the_queue.size();
        
        if (size != packets_in_queue || size < 0) {
            System.err.println("Internal error in PackeQueue::size()");
            System.exit(1);
        }
        
        return size;
    }
}
/*
 * vim: et ts=4 sw=4
 */

