/*
 * (c) Copyright 2004 Mauro Tortonesi - ALL RIGHTS RESERVED
 * 
 * This file is a driver for the JavaBluetooth Stack that interfaces
 * JavaBluetooth with Bluez.
 * 
 * The JavaBluetooth Stack is free software; you can redistribute it 
 * and/or modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 * 
 * The JavaBluetooth Stack is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 */

package org.javabluetooth.stack.sco;

import java.io.*;
import java.util.*;
import javax.bluetooth.RemoteDevice;
import org.javabluetooth.util.Debug;
import org.javabluetooth.util.PacketQueue;
import org.javabluetooth.util.BluetoothAddress;
import org.javabluetooth.util.BluetoothAddressException;
import org.javabluetooth.stack.sco.SCOException;

/** 
 * BluezSCOConnection is an interface to the Bluez protocol stack for the exploiting
 * of SCO channels.
 * @author Mauro Tortonesi
 */
public class BluezSCOConnection implements SCOConnection, Runnable {
    private PacketQueue tx_queue;
    private PacketQueue rx_queue;
    private Object rx_queue_empty;
    private RemoteDevice remoteEndPoint;

    static {
        /* Load libBluezSCOConnection.so (which must be in one of the directories
         * in /etc/ld.so.conf or in the LD_LIBRARY_PATH environment variable). */
        System.loadLibrary("BluezSCOConnection");
    }

    /* native method, implemented in libBluezSCOConnection.c */
    private native int startSCO(String bd_addr, PacketQueue tx_queue);
    
    /**
     * BluezSCOConnection constructor.
     */
    public BluezSCOConnection(String InterfaceName, RemoteDevice remoteDevice) {
        /* For the moment, the InterfaceName parameter is ignored. In future, 
         * developers will be able to choose the Bluez interface, e.g., "hciX",
         * they wish to use. */
        tx_queue = new PacketQueue();
        rx_queue = new PacketQueue();
        rx_queue_empty = new Object();
        remoteEndPoint = remoteDevice;
        Thread t = new Thread(this);
        t.start();
        // reset();
    }
    
    public void run() {
        Debug.println(1, "inside BluezSCOConnection::run()");
        String jsr82_addr = remoteEndPoint.getBluetoothAddress();
        String bd_addr = BluetoothAddress.JSR82ToStandardRepr(jsr82_addr);
        int ret = startSCO(bd_addr, tx_queue);
        Debug.println(1, "exiting from BluezSCOConnection::run(), startSCO returned: " + ret);
    }
    
    /**
     * Queues data for transmission over a SCO Channel. The bytes are send
     * via a low-level SCO socket interface to the Bluez protocol stack.
     * @throws SCOException 
     */
    public synchronized void send(byte[] data) {
        Debug.println(1, "SCO: Queuing Packet for TX: ", data);
        tx_queue.insert(data);
    }
    
    /**
     * Reads incoming data. If there is no data in the internal rx_queue, 
     * block until some data is available.
     */
    public synchronized int receive(byte[] inBuf) {
        Debug.println(1, "in BluezSCOConnection::receive()");
        
        /* if rx_queue is empty, block waiting for incoming data */
        synchronized(rx_queue_empty) {
            if (rx_queue.isEmpty()) {
                Debug.println(1, "in BluezSCOConnection::receive() rx_queue is empty: waiting...");
                try {
                    rx_queue_empty.wait();
                }
                catch (InterruptedException e) { }
            }
        }
        
        /* copy data from the rx_queue to inBuf */
        byte[] packet = (byte[])rx_queue.extract();
        if (packet != null) {
            int length = inBuf.length;
            if (packet.length > length) {
                Debug.println(1, "SCO: lost " + (packet.length - length) + " bytes of data");
            } else {
                length = packet.length;
            }
            System.arraycopy(packet, 0, inBuf, 0, length);
            return length;
        } else {
            Debug.println(1, "in BluezSCOConnection::receive() packet is null");
        }
        
        Debug.println(1, "out from BluezSCOConnection::receive()");

        return 0;
    }
    
    /* no need to use "synchronized" for this method */
    private void receiveData(byte[] data) {
        rx_queue.insert(data);
        
        /* if rx_queue was empty, notify it is not anymore */
        synchronized(rx_queue_empty) {
            if (!rx_queue.isEmpty()) {
                Debug.println(1, "in BluezSCOConnection::receiveData() rx_queue is not empty: notifying...");
                rx_queue_empty.notify();
            }
        }        
    }

    public void close() {
        /* TODO: implement a way to close the SCO socket */
    }
    
    public boolean ready() throws IOException {
        return true;
    }

    public RemoteDevice getRemoteDevice() {
        return remoteEndPoint;
    }

    public int getTransmitMTU() throws IOException {
        throw new RuntimeException("This implementation of SCOConnection " +
                                   "does not support getTransmitMTU().");
    }
    
    public int getReceiveMTU() throws IOException {
        throw new RuntimeException("This implementation of SCOConnection " +
                                   "does not support getReceiveMTU().");
    }
}
/* 
 * vim: et ts=4 sw=4 
 */

